# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import HagerZhangLineSearch
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot

import numpy as onp

from sklearn import datasets


class HagerZhangLinesearchTest(test_util.JaxoptTestCase):

  def _check_conditions_satisfied(
      self,
      c1,
      c2,
      stepsize,
      initial_value,
      initial_grad,
      final_state):
    self.assertTrue(jnp.all(final_state.done))
    self.assertFalse(jnp.any(final_state.failed))

    descent_direction = tree_scalar_mul(-1, initial_grad)
    sufficient_decrease = jnp.all(
        final_state.value <= initial_value +
        c1 * stepsize * tree_vdot(final_state.grad, descent_direction))
    self.assertTrue(sufficient_decrease)

    new_gd_vdot = tree_vdot(final_state.grad, descent_direction)
    gd_vdot = tree_vdot(initial_grad, descent_direction)
    curvature = jnp.all(new_gd_vdot >= c2 * gd_vdot)
    self.assertTrue(curvature)

  def test_hager_zhang_linesearch(self):
    x, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2,
        n_informative=3, random_state=0)
    data = (x, y)
    fun = objective.binary_logreg

    rng = onp.random.RandomState(0)
    w_init = rng.randn(x.shape[1])
    initial_grad = jax.grad(fun)(w_init, data=data)
    initial_value = fun(w_init, data=data)

    # Manual loop.
    ls = HagerZhangLineSearch(fun=fun)
    stepsize = 1.0
    state = ls.init_state(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    stepsize, state = ls.update(stepsize=stepsize, state=state, params=w_init,
                                fun_kwargs={"data": data})

    # Call to run.
    ls = HagerZhangLineSearch(fun=fun, maxiter=20)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self._check_conditions_satisfied(
        ls.c1, ls.c2, stepsize, initial_value, initial_grad, state)

    # Call to run with value_and_grad=True.
    ls = HagerZhangLineSearch(fun=jax.value_and_grad(fun),
                              maxiter=20,
                              value_and_grad=True)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self._check_conditions_satisfied(
        ls.c1, ls.c2, stepsize, initial_value, initial_grad, state)

    # Failed linesearch (high c1 ensures convergence condition is not met).
    ls = HagerZhangLineSearch(fun=fun, maxiter=20, c1=2.)
    _, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self.assertTrue(jnp.all(state.failed))
    self.assertFalse(jnp.any(state.done))

  @parameterized.product(val=[onp.inf, onp.nan])
  def test_hager_zhang_linesearch_non_finite(self, val):

    def fun(x):
      result = jnp.where(x > 4., val, (x - 2)**2)
      grad = jnp.where(x > 4., onp.nan, 2 * (x - 2.))
      return result, grad
    x_init = -0.001

    ls = HagerZhangLineSearch(fun=fun, value_and_grad=True, jit=False)
    stepsize = 1.25
    state = ls.init_state(init_stepsize=1.25, params=x_init)

    stepsize, state = ls.update(stepsize=stepsize, state=state, params=x_init)
    # Should work around the Nan/Inf regions and provide a reasonable step size.
    self.assertTrue(state.done)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of gradient descent in JAX."""

from typing import Any
from typing import NamedTuple

from dataclasses import dataclass

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.proximal_gradient import ProximalGradient, ProxGradState


@dataclass(eq=False)
class GradientDescent(ProximalGradient):
  """Gradient Descent solver.

  Attributes:
    fun: a smooth function of the form ``fun(parameters, *args, **kwargs)``,
      where ``parameters`` are the model parameters w.r.t. which we minimize
      the function and the rest are fixed auxiliary parameters.
    value_and_grad: whether ``fun`` just returns the value (False) or both
      the value and gradient (True).
    has_aux: whether ``fun`` outputs auxiliary data or not.
      If ``has_aux`` is False, ``fun`` is expected to be
        scalar-valued.
      If ``has_aux`` is True, then we have one of the following
        two cases.
      If ``value_and_grad`` is False, the output should be
      ``value, aux = fun(...)``.
      If ``value_and_grad == True``, the output should be
      ``(value, aux), grad = fun(...)``.
      At each iteration of the algorithm, the auxiliary outputs are stored
        in ``state.aux``.

    stepsize: a stepsize to use (if <= 0, use backtracking line search), or a
      callable specifying the **positive** stepsize to use at each iteration.
    maxiter: maximum number of proximal gradient descent iterations.
    maxls: maximum number of iterations to use in the line search.
    tol: tolerance to use.

    acceleration: whether to use acceleration (also known as FISTA) or not.
    verbose: whether to print information on every iteration or not.

    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.

    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").
    """

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> ProxGradState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      state
    """
    return super().init_state(init_params, None, *args, **kwargs)

  def update(
      self, params: Any, state: ProxGradState, *args, **kwargs
  ) -> base.OptStep:
    """Performs one iteration of gradient descent.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      (params, state)
    """
    return super().update(params, state, None, *args, **kwargs)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    return self._grad_fun(params, *args, **kwargs)

  def __post_init__(self):
    super().__post_init__()
    self.reference_signature = self.fun



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import BlockCoordinateDescent
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt import prox
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp

from sklearn import datasets
from sklearn import preprocessing


class BlockCoordinateDescentTest(test_util.JaxoptTestCase):

  def test_lasso_manual_loop(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)

    # Setup parameters.
    fun = objective.least_squares  # fun(params, data)
    l2reg = 10.0
    data = (X, y)

    # Initialize.
    params = jnp.zeros(X.shape[1])
    bcd = BlockCoordinateDescent(fun=fun, block_prox=prox.prox_lasso)
    state = bcd.init_state(init_params=params,
                           hyperparams_prox=l2reg,
                           data=data)
    # Optimization loop.
    for _ in range(30):
      params, state = bcd.update(params=params, state=state,
                                 hyperparams_prox=l2reg, data=data)

    # Check optimality conditions.
    self.assertLess(state.error, 0.5)

  def test_lasso(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)

    # Set up parameters.
    fun = objective.least_squares  # fun(params, data)
    l2reg = 10.0
    data = (X, y)
    w_init = jnp.zeros(X.shape[1])

    # Run solver.
    bcd = BlockCoordinateDescent(fun=fun,
                                 block_prox=prox.prox_lasso,
                                 maxiter=150)
    sol = bcd.run(init_params=w_init, hyperparams_prox=l2reg, data=data)

    # Check optimality conditions.
    self.assertLess(sol.state.error, 0.01)

    # Check against sklearn.
    w_skl = test_util.lasso_skl(X, y, l2reg)
    self.assertArraysAllClose(sol.params, w_skl, atol=1e-2)

  def test_elastic_net(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)

    # Set up parameters.
    fun = objective.least_squares  # fun(params, data)
    hyperparams_prox = (2.0, 0.8)
    data = (X, y)
    w_init = jnp.zeros(X.shape[1])

    # Run solver.
    bcd = BlockCoordinateDescent(fun=fun,
                                 block_prox=prox.prox_elastic_net,
                                 maxiter=100)
    sol = bcd.run(init_params=w_init,
                  hyperparams_prox=hyperparams_prox,
                  data=data)

    # Check optimality conditions.
    self.assertLess(sol.state.error, 0.01)

    # Check against sklearn.
    w_skl = test_util.enet_skl(X, y, hyperparams_prox)
    self.assertArraysAllClose(sol.params, w_skl, atol=1e-2)

  def test_multitask_reg(self):
    # Generate data.
    rng = onp.random.RandomState(0)
    n_samples, n_features, n_tasks = 50, 10, 3
    X = rng.randn(n_samples, n_features)
    W = rng.randn(n_features, n_tasks)
    Y = jnp.dot(X, W) + rng.randn(n_samples, n_tasks)

    # Set up parameters.
    fun = objective.least_squares  # fun(params, data)
    block_prox = prox.prox_group_lasso
    l2reg = 1e-1
    W_init = jnp.zeros((n_features, n_tasks))
    data = (X, Y)

    # Run solver.
    bcd = BlockCoordinateDescent(fun=fun, block_prox=block_prox,
                                 maxiter=1000, tol=1e-3)
    sol = bcd.run(init_params=W_init, hyperparams_prox=l2reg, data=data)

    # Check optimality conditions.
    self.assertLess(sol.state.error, 0.01)

    # Compare against sklearn.
    W_skl = test_util.multitask_lasso_skl(X, Y, l2reg * n_tasks)
    self.assertArraysAllClose(sol.params, W_skl, atol=1e-1)

  @parameterized.product(multiclass=[True, False], penalty=["l1", "l2"])
  def test_logreg(self, multiclass, penalty):
    # Generate data.
    if multiclass:
      n_samples, n_features, n_classes = 20, 5, 3
      W_init = jnp.zeros((n_features, n_classes))
    else:
      n_samples, n_features, n_classes = 20, 5, 2
      W_init = jnp.zeros(n_features)

    X, y = datasets.make_classification(n_samples=n_samples,
                                        n_features=n_features,
                                        n_informative=3, n_classes=n_classes,
                                        random_state=0)
    data = (X, y)

    # Set up parameters.
    if penalty == "l1":
      block_prox = prox.prox_lasso
    else:
      block_prox = prox.prox_ridge

    if multiclass:
      fun = objective.multiclass_logreg
    else:
      fun = objective.binary_logreg

    l2reg = 1e-2

    # Run solver.
    bcd = BlockCoordinateDescent(fun=fun, block_prox=block_prox,
                                 maxiter=3500, tol=1e-5)
    sol = bcd.run(W_init, hyperparams_prox=l2reg, data=data)

    # Check optimality conditions.
    self.assertLess(sol.state.error, 0.01)

    if not (multiclass and penalty == "l1"):

      # Compare against sklearn (it does not support multiclass + l1).
      W_skl = test_util.logreg_skl(X, y, l2reg, penalty=penalty,
                                   multiclass=multiclass)
      self.assertArraysAllClose(sol.params, W_skl, atol=1e-2)

      # Check differentiation.
      jac_num = test_util.logreg_skl_jac(X, y, l2reg, eps=1e-4,
                                         penalty=penalty, multiclass=multiclass)

      # By autodiff.
      bcd = BlockCoordinateDescent(fun=fun,
                                   block_prox=block_prox,
                                   maxiter=10000, tol=1e-6,
                                   implicit_diff=False)
      def wrapper(hyperparams_prox):
        return bcd.run(init_params=W_init,
                       hyperparams_prox=hyperparams_prox,
                       data=data).params
      jac = jax.jacfwd(wrapper)(l2reg)
      self.assertAllClose(jac_num, jac, atol=5e-1)

      # By implicit diff.
      bcd = BlockCoordinateDescent(fun=fun, block_prox=block_prox,
                                   maxiter=3500, tol=1e-5, implicit_diff=True)
      def wrapper(hyperparams_prox):
        return bcd.run(W_skl, hyperparams_prox, data).params

      jac = jax.jacrev(wrapper)(l2reg)
      self.assertAllClose(jac_num, jac, atol=1e-1)

      # Test again with kwargs.
      def wrapper(hyperparams_prox):
        return bcd.run(W_skl, hyperparams_prox=hyperparams_prox, data=data).params

      jac = jax.jacrev(wrapper)(l2reg)
      self.assertAllClose(jac_num, jac, atol=1e-1)

  def test_multiclass_linear_svm(self):
    # Generate data.
    n_samples, n_classes = 20, 3
    X, y = datasets.make_classification(n_samples=n_samples, n_features=5,
                                        n_informative=3, n_classes=n_classes,
                                        random_state=0)
    Y = preprocessing.LabelBinarizer().fit_transform(y)
    Y = jnp.array(Y)

    # Set up parameters.
    block_prox = prox.make_prox_from_projection(projection.projection_simplex)
    fun = objective.multiclass_linear_svm_dual
    data = (X, Y)
    l2reg = 1000.0
    beta_init = jnp.ones((n_samples, n_classes)) / n_classes

    # Run solver.
    bcd = BlockCoordinateDescent(fun=fun, block_prox=block_prox,
                                 maxiter=3500, tol=1e-5)
    sol = bcd.run(beta_init, hyperparams_prox=None, l2reg=l2reg, data=data)

    # Check optimality conditions.
    self.assertLess(sol.state.error, 0.01)

    # Compare against sklearn.
    W_skl = test_util.multiclass_linear_svm_skl(X, y, l2reg)
    W_fit = jnp.dot(X.T, (Y - sol.params)) / l2reg
    self.assertArraysAllClose(W_fit, W_skl, atol=1e-3)

  def test_raise_exception(self):
    self.assertRaises(AttributeError, BlockCoordinateDescent,
                      fun=objective.multiclass_logreg_with_intercept,
                      block_prox=prox.prox_lasso)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



"""Custom FeaturePytree class.

Implement the ``FeaturePytree`` class that stores inputs as a dictionary of arrays.
These objects can be used as inputs to NeMoS models. The fitted coefficients will be stored as
dictionaries with matching keys, facilitating interpretation of the model output.
"""

from collections import UserDict

import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from .tree_utils import pytree_map_and_reduce


@register_pytree_node_class
class FeaturePytree(UserDict):
    """Pytree to represent GLM features.

    This object is essentially a dictionary with strings as its keys and
    n-dimensional array-like objects as its values. FeaturePytree objects can
    only have a depth of 1, and we allow joint slicing.

    This is intended to be used with jax.tree_util.tree_map and similar functionality.
    """

    def __init__(self, **kwargs):
        self._num_time_points = None
        super().__init__(kwargs)

    @property
    def shape(self):
        """Return the sample points, i.e. the shared dimension."""
        # Every value is an array with the same number of time points, so
        # that's the only shared shape.
        return (self._num_time_points,)

    def __len__(self):
        """Return the number of sample points."""
        # Same logic as shape
        return self._num_time_points

    @property
    def ndim(self):
        """Force slicing only on the time axis."""
        # You can slice into a FeaturePytree along one dimension (time)
        return 1

    def __setitem__(self, key, value):
        """Check arrays and set the values."""
        # All keys are strings
        if not isinstance(key, str):
            raise ValueError("Keys must be strings!")
        # All values are array-like and must have time points along the first
        # dimension, so double-check that's the case here.
        try:
            value.shape[0]
        except (AttributeError, IndexError):
            raise ValueError("All values must be arrays of at least 1 dimension!")
        # For the first value, we must set how time points this FeaturePytree
        # has
        if self._num_time_points is None:
            self._num_time_points = value.shape[0]
        # For all subsequent values, ensure they have the proper number of time points
        else:
            if self._num_time_points != value.shape[0]:
                raise ValueError(
                    f"All arrays must have same number of time points, {self._num_time_points}"
                )
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """Getitem using data dictionary keys."""
        # We can index into either the features, which have strs as their
        # keys...
        if isinstance(key, str):
            return self.data[key]
        # Or the time dimension
        else:
            return jax.tree_util.tree_map(lambda x: x[key], self)

    def __repr__(self):
        """Represent a FeaturePytree."""
        # Show the shape and data type of each array
        repr = [f"{k}: shape {v.shape}, dtype {v.dtype}" for k, v in self.data.items()]
        return "\n".join(repr)

    def __eq__(self, other):
        """Equality over a tree of arrays."""
        # if structure is different, pytree_map_and_reduce will return a ValueError
        if jax.tree_util.tree_structure(self) != jax.tree_util.tree_structure(other):
            return False
        return pytree_map_and_reduce(
            lambda x, y: jnp.array_equal(x, y), all, self, other
        )

    def tree_flatten(self):
        """Flatten data."""
        return jax.tree_util.tree_flatten(self.data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Unflatten the data dict and return the class."""
        try:
            return cls(**jax.tree_util.tree_unflatten(aux_data, children))
        except ValueError:
            return jax.tree_util.tree_unflatten(aux_data, children)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import jax
from jax.nn import softmax
from jax.nn import softplus
import jax.numpy as jnp
from jax.scipy.special import expit as sigmoid
from jax.scipy.special import logsumexp
from nemos.third_party.jaxopt.jaxopt import loss
from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt._src import test_util
import numpy as onp


def one_hot_argmax(inputs: jnp.ndarray) -> jnp.ndarray:
  """An argmax one-hot function for arbitrary shapes."""
  inputs_flat = jnp.reshape(inputs, (-1))
  flat_one_hot = jax.nn.one_hot(jnp.argmax(inputs_flat), inputs_flat.shape[0])
  return jnp.reshape(flat_one_hot, inputs.shape)


class LossTest(test_util.JaxoptTestCase):

  def _test_binary_loss_function(self, loss_fun, inv_link_fun, reference_impl):
    # Check that loss is zero when the weight goes to the correct label.
    loss_val = loss_fun(0, -1e5)
    self.assertEqual(loss_val, 0)
    loss_val = loss_fun(1, 1e5)
    self.assertEqual(loss_val, 0)

    # Check that gradient has the expected form.
    logit = 1.2
    grad = jax.grad(loss_fun, argnums=1)(0, logit)
    self.assertArraysAllClose(grad, inv_link_fun(logit))
    grad = jax.grad(loss_fun, argnums=1)(1, logit)
    self.assertArraysAllClose(grad, inv_link_fun(logit) - 1)

    # Check that vmap works as expected.
    logits = jnp.array([1.2, -0.3, 0.7])
    labels = jnp.array([0, 0, 1])
    losses = jax.vmap(loss_fun)(labels, logits)
    losses2 = jnp.array([loss_fun(labels[0], logits[0]),
                         loss_fun(labels[1], logits[1]),
                         loss_fun(labels[2], logits[2])])
    self.assertArraysAllClose(losses, losses2)

    # Compare against intuitive reference implementation.
    rng = onp.random.RandomState(0)
    for _ in range(10):
      logit = jnp.array(rng.randn(), dtype=jnp.float32)
      label = jnp.array(rng.randint(2))
      loss_val = loss_fun(label, logit)
      expected = reference_impl(label, logit)
      self.assertAllClose(loss_val, expected)

    # Check that correct value is obtained for large inputs.
    loss_val = loss_fun(0, -1e9)
    self.assertAllClose(loss_val, 0.0)
    loss_val = loss_fun(0, 1e9)
    self.assertAllClose(loss_val, 1e9)

    # Check that loss behaves correctly with infinite input.
    loss_val = loss_fun(1, jnp.inf)
    self.assertEqual(loss_val, 0.0)
    loss_val = loss_fun(0, -jnp.inf)
    self.assertEqual(loss_val, 0.0)
    loss_val = loss_fun(1, -jnp.inf)
    self.assertEqual(loss_val, jnp.inf)
    loss_val = loss_fun(0, jnp.inf)
    self.assertEqual(loss_val, jnp.inf)

    # Check that gradient has the expected form for large inputs.
    logit = 1e9
    grad = jax.grad(loss_fun, argnums=1)(0, logit)
    self.assertArraysAllClose(grad, inv_link_fun(logit))
    grad = jax.grad(loss_fun, argnums=1)(1, logit)
    self.assertArraysAllClose(grad, inv_link_fun(logit) - 1)

  def test_binary_logistic_loss(self):
    reference_impl = lambda label, logit: softplus(logit) - label * logit
    self._test_binary_loss_function(loss.binary_logistic_loss, sigmoid,
                                    reference_impl)

  def test_binary_sparsemax_loss(self):
    def reference_impl(label: int, logit: float) -> float:
      scores = -(2*label-1)*logit
      if scores <= -1.0:
        return 0.0
      elif scores >= 1.0:
        return scores
      else:
        return (scores + 1.0) ** 2 / 4

    self._test_binary_loss_function(
        loss.binary_sparsemax_loss, loss.sparse_sigmoid, reference_impl
    )

  def test_binary_hinge_loss(self):
    def reference_impl(label: int, logit: float) -> float:
      return jax.nn.relu(1 - logit * (2.0 * label - 1.0))
    self._test_binary_loss_function(loss.binary_hinge_loss, jnp.sign,
                                    reference_impl)

  def test_perceptron_loss(self):
    def reference_impl(label: int, logit: float) -> float:
      return jax.nn.relu(- logit * (2.0 * label - 1.0))
    self._test_binary_loss_function(loss.binary_perceptron_loss, jnp.sign,
                                    reference_impl) 

  def _test_multiclass_loss_function(
      self, loss_fun, inv_link_fun, reference_impl, large_inputs_behavior=True,
      incorrect_label_infty=True
  ):
    # Check that loss is zero when all weights goes to the correct label.
    loss_val = loss_fun(0, jnp.array([1e5, 0, 0]))
    self.assertEqual(loss_val, 0)

    # Check that gradient has the expected form.
    logits = jnp.array([1.2, 0.3, 2.3])
    grad = jax.grad(loss_fun, argnums=1)(0, logits)
    self.assertArraysAllClose(grad, inv_link_fun(logits) - jnp.array([1, 0, 0]))

    # Check that vmap works as expected.
    logits = jnp.array([[1.3, 2.5],
                        [1.7, -0.3],
                        [0.2, 1.2]])
    labels = jnp.array([0, 0, 1])
    losses = jax.vmap(loss_fun)(labels, logits)
    losses2 = jnp.array([loss_fun(labels[0], logits[0]),
                         loss_fun(labels[1], logits[1]),
                         loss_fun(labels[2], logits[2])])
    self.assertArraysAllClose(losses, losses2)

    # Compare against intuitive reference implementation.
    rng = onp.random.RandomState(0)
    for _ in range(10):
      n_classes = rng.randint(8, 16, size=())
      logits = jnp.array(rng.randn(n_classes), dtype=jnp.float32)
      label = jnp.array(rng.randint(n_classes))
      loss_val = loss_fun(label, logits)
      expected = reference_impl(label, logits)
      self.assertAllClose(loss_val, expected)

    # Check that correct value is obtained for large inputs.
    if large_inputs_behavior:
      loss_val = loss_fun(0, jnp.array([1e9, 1e9]))
      expected = loss_fun(0, jnp.array([1., 1.]))
      self.assertAllClose(loss_val, expected)

    # Check that -inf for incorrect label has no impact.
    if incorrect_label_infty:
      loss_val = loss_fun(0, jnp.array([0.0, 0.0, -jnp.inf]))
      expected = loss_fun(0, jnp.array([0.0, 0.0]))
      self.assertAllClose(loss_val, expected)
    # Check that -inf for correct label results in infinite loss.
    loss_val = loss_fun(0, jnp.array([-jnp.inf, 0.0, 0.0]))
    self.assertEqual(loss_val, jnp.inf)
    # Check that nan is contagious.
    loss_val = loss_fun(0, jnp.array([0.0, 0.0, jnp.nan]))
    self.assertTrue(jnp.isnan(loss_val))
    loss_val = loss_fun(0, jnp.array([jnp.nan, 0.0, 0.0]))
    self.assertTrue(jnp.isnan(loss_val))

    # Check that gradient has the expected form for large inputs.
    logits = 1e9 * jnp.array([1.2, 0.3, 2.3])
    grad = jax.grad(loss_fun, argnums=1)(0, logits)
    self.assertArraysAllClose(grad, inv_link_fun(logits) - jnp.array([1, 0, 0]))

    # Check that gradient has the expected form with -inf inputs.
    logits = jnp.array([1.2, -jnp.inf, 2.3])
    grad = jax.grad(loss_fun, argnums=1)(0, logits)
    self.assertArraysAllClose(grad, inv_link_fun(logits) - jnp.array([1, 0, 0]))

  def test_multiclass_logistic_loss(self):
    def reference_impl(label, logits):
      n_classes = logits.shape[0]
      one_hot = jax.nn.one_hot(label, n_classes)
      return logsumexp(logits) - jnp.dot(logits, one_hot)

    self._test_multiclass_loss_function(loss.multiclass_logistic_loss, softmax,
                                        reference_impl)

  def test_multiclass_sparsemax_loss(self):
    def reference_impl(label, scores):
      n_classes = scores.shape[0]
      one_hot = jax.nn.one_hot(label, n_classes)
      proba = projection.projection_simplex(scores)
      cumulant = 0.5 + jnp.dot(proba, scores - 0.5 * proba)
      return cumulant - jnp.dot(scores, one_hot)

    self._test_multiclass_loss_function(loss.multiclass_sparsemax_loss,
                                        projection.projection_simplex,
                                        reference_impl)

  def test_multiclass_hinge_loss(self):
    def reference_impl(label, scores):
      one_hot_label = jax.nn.one_hot(label, scores.shape[-1])
      return jnp.max(scores + 1.0 - one_hot_label) - scores[label]
    def inv_link_fun(scores):
      return jax.nn.one_hot(jnp.argmax(scores), scores.shape[-1])

    self._test_multiclass_loss_function(loss.multiclass_hinge_loss,
                                        inv_link_fun,
                                        reference_impl,
                                        large_inputs_behavior=False,
                                        incorrect_label_infty=False)

  def test_multiclass_perceptron_loss(self):
    def reference_impl(label, scores):
      return jnp.max(scores) - scores[label]
    def inv_link_fun(scores):
      return jax.nn.one_hot(jnp.argmax(scores), scores.shape[-1])

    self._test_multiclass_loss_function(loss.multiclass_perceptron_loss,
                                        inv_link_fun,
                                        reference_impl,
                                        incorrect_label_infty=False)

  def test_huber(self):
    self.assertAllClose(0.0, loss.huber_loss(0, 0, .1))
    self.assertAllClose(0.0, loss.huber_loss(1, 1, .1))
    self.assertAllClose(.1 * (1.0- .5 * .1), loss.huber_loss(4, 3, .1))
    self.assertAllClose(0.125, loss.huber_loss(0, .5))

  def test_fenchel_young_reg(self):
    # Checks the behavior of the Fenchel-Young loss.
    fy_loss = loss.make_fenchel_young_loss(logsumexp)
    rng = jax.random.PRNGKey(0)
    rngs = jax.random.split(rng, 2)
    theta_true = jax.random.uniform(rngs[0], (8, 5))
    y_true = jax.vmap(jax.nn.softmax)(theta_true)
    theta_random = jax.random.uniform(rngs[1], (8, 5))
    y_random = jax.vmap(jax.nn.softmax)(theta_random)
    fy_min = jax.vmap(fy_loss)(y_true, theta_true)
    fy_random = jax.vmap(fy_loss)(y_true, theta_random)
    # Checks that the loss is minimized for true value of the parameters.
    self.assertGreater(fy_random[0], fy_min[0])
    self.assertGreater(jnp.mean(fy_random), jnp.mean(fy_min))
    grad_random = jax.vmap(jax.grad(fy_loss, argnums=1))(y_true, theta_random)
    # Checks that the gradient of the loss takes the correct form.
    self.assertArraysAllClose(grad_random, y_random - y_true)
    y_one_hot = jax.vmap(one_hot_argmax)(theta_true)
    int_one_hot = jnp.where(y_one_hot == 1.)[1]
    loss_one_hot = jax.vmap(fy_loss)(y_one_hot, theta_random)
    log_loss = jax.vmap(loss.multiclass_logistic_loss)(int_one_hot,
                                                       theta_random)
    # Checks that the FY loss associated to logsumexp is correct.
    self.assertArraysAllClose(loss_one_hot, log_loss)


if __name__ == '__main__':
  absltest.main()



"""Observation model classes for GLMs."""

import abc
from typing import Callable, Literal, Union

import jax
import jax.numpy as jnp
from numpy.typing import NDArray

from . import utils
from .base_class import Base

__all__ = [
    "PoissonObservations",
    "GammaObservations",
    "BernoulliObservations",
    "NegativeBinomialObservations",
]


def __dir__():
    return __all__


class Observations(Base, abc.ABC):
    """
    Abstract observation model class for neural data processing.

    This is an abstract base class used to implement observation models for neural data.
    Specific observation models that inherit from this class should define their versions
    of the abstract methods such as :meth:`~nemos.observation_models.Observations.log_likelihood`,
    :meth:`~nemos.observation_models.Observations.sample_generator`, and
    :meth:`~nemos.observation_models.Observations.deviance`.

    See Also
    --------
    :class:`~nemos.observation_models.PoissonObservations`
        A specific implementation of an observation model using the Poisson distribution.
    :class:`~nemos.observation_models.GammaObservations`
        A specific implementation of an observation model using the Gamma distribution.
    :class:`~nemos.observation_models.BernoulliObservations`
        A specific implementation of an observation model using the Bernoulli distribution.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.scale = 1.0

    def __repr__(self):
        return utils.format_repr(self)

    @property
    @abc.abstractmethod
    def default_inverse_link_function(self):
        pass

    @property
    def scale(self):
        """Getter for the scale parameter of the model."""
        return self._scale

    @scale.setter
    def scale(self, value: Union[int, float, jnp.ndarray]):
        """Setter for the scale parameter of the model."""
        try:
            self._scale = float(value)
        except Exception:
            raise ValueError("The `scale` parameter must be of numeric type.")

    @abc.abstractmethod
    def _negative_log_likelihood(
        self, y, predicted_rate, aggregate_sample_scores: Callable = jnp.mean
    ):
        r"""Compute the observation model negative log-likelihood.

        This computes the negative log-likelihood of the predicted rates
        for the observed neural activity up to a constant.

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons)..
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons)..

        Returns
        -------
        :
            The negative log-likehood. Shape (1,).
        """
        pass

    @abc.abstractmethod
    def log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the observation model log-likelihood.

        This computes the log-likelihood of the predicted rates
        for the observed neural activity including the normalization constant

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        scale :
            The scale parameter of the model
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The log-likehood. Shape (1,).
        """
        pass

    def likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the observation model likelihood.

        This computes the likelihood of the predicted rates
        for the observed neural activity including the normalization constant.

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        scale :
            The scale parameter of the model
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The likelihood. Shape (1,).
        """
        return jnp.exp(
            self.log_likelihood(y, predicted_rate, scale, aggregate_sample_scores)
        )

    @abc.abstractmethod
    def sample_generator(
        self,
        key: jax.Array,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        """
        Sample from the estimated distribution.

        This method generates random numbers from the desired distribution based on the given
        `predicted_rate`.

        Parameters
        ----------
        key :
            Random key used for the generation of random numbers in JAX.
        predicted_rate :
            Expected rate of the distribution. Shape (n_time_bins, ), or (n_time_bins, n_neurons)..
        scale:
            Scale parameter for the distribution.

        Returns
        -------
        :
            Random numbers generated from the observation model with `predicted_rate`.
        """
        pass

    @abc.abstractmethod
    def deviance(
        self,
        spike_counts: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ):
        r"""Compute the residual deviance for the observation model.

        Parameters
        ----------
        spike_counts:
            The spike counts. Shape ``(n_time_bins, )`` or ``(n_time_bins, n_neurons)`` for population models.
        predicted_rate:
            The predicted firing rates. Shape ``(n_time_bins, )`` or ``(n_time_bins, n_neurons)`` for population models.
        scale:
            Scale parameter of the model.

        Returns
        -------
        :
            The residual deviance of the model.
        """
        pass

    @abc.abstractmethod
    def estimate_scale(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        dof_resid: Union[float, jnp.ndarray],
    ) -> Union[float, jnp.ndarray]:
        r"""Estimate the scale parameter for the model.

        This method estimates the scale parameter, often denoted as :math:`\phi`, which determines the dispersion
        of an exponential family distribution. The probability density function (pdf) for such a distribution
        is generally expressed as
        :math:`f(x; \theta, \phi) \propto \exp \left(a(\phi)\left(  y\theta - \mathcal{k}(\theta) \right)\right)`.

        The relationship between variance and the scale parameter is given by:

        .. math::
           \text{var}(Y) = \frac{V(\mu)}{a(\phi)}.

        The scale parameter, :math:`\phi`, is necessary for capturing the variance of the data accurately.

        Parameters
        ----------
        y :
            Observed activity.
        predicted_rate :
            The predicted rate values.
        dof_resid :
            The DOF of the residual.
        """
        pass

    def pseudo_r2(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        score_type: Literal[
            "pseudo-r2-McFadden", "pseudo-r2-Cohen"
        ] = "pseudo-r2-McFadden",
        scale: Union[float, jnp.ndarray, NDArray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Pseudo-:math:`R^2` calculation for a GLM.

        Compute the pseudo-:math:`R^2` metric for the GLM, as defined by McFadden et al. [2]_
        or by Cohen et al. [3]_.

        This metric evaluates the goodness-of-fit of the model relative to a null (baseline) model that assumes a
        constant mean for the observations. While the pseudo-:math:`R^2` is bounded between 0 and 1 for the
        training set, it can yield negative values on out-of-sample data, indicating potential over-fitting.

        Parameters
        ----------
        y:
            The neural activity. Expected shape: ``(n_time_bins, )``
        predicted_rate:
            The mean neural activity. Expected shape: ``(n_time_bins, )``
        score_type:
            The pseudo-:math:`R^2` type.
        scale:
            The scale parameter of the model.

        Returns
        -------
        :
            The pseudo-:math:`R^2` of the model. A value closer to 1 indicates a better model fit,
            whereas a value closer to 0 suggests that the model doesn't improve much over the null model.

        Notes
        -----
        - The McFadden pseudo-:math:`R^2` is given by:

          .. math::
                R^2_{\text{mcf}} = 1 - \frac{\log(L_{M})}{\log(L_0)}.

          *Equivalent to statsmodels*
          `GLMResults.pseudo_rsquared(kind='mcf') <https://www.statsmodels.org/dev/generated/statsmodels.genmod.
          generalized_linear_model.GLMResults.pseudo_rsquared.html>`_ .

        - The Cohen pseudo-:math:`R^2` is given by:

          .. math::
               \begin{aligned}
               R^2_{\text{Cohen}} &= \frac{D_0 - D_M}{D_0} \\\
               &= 1 - \frac{\log(L_s) - \log(L_M)}{\log(L_s)-\log(L_0)},
               \end{aligned}

          where :math:`L_M`, :math:`L_0` and :math:`L_s` are the likelihood of the fitted model, the null model (a
          model with only the intercept term), and the saturated model (a model with one parameter per
          sample, i.e. the maximum value that the likelihood could possibly achieve). :math:`D_M` and :math:`D_0` are
          the model and the null deviance, :math:`D_i = -2 \left[ \log(L_s) - \log(L_i) \right]` for :math:`i=M,0`.

        References
        ----------
        .. [2] McFadden D (1979). Quantitative methods for analysing travel behavior of individuals: Some recent
               developments. In D. A. Hensher & P. R. Stopher (Eds.), *Behavioural travel modelling* (pp. 279-318).
               London: Croom Helm.

        .. [3] Jacob Cohen, Patricia Cohen, Steven G. West, Leona S. Aiken.
               *Applied Multiple Regression/Correlation Analysis for the Behavioral Sciences*.
               3rd edition. Routledge, 2002. p.502. ISBN 978-0-8058-2223-6. (May 2012)
        """
        if score_type == "pseudo-r2-McFadden":
            pseudo_r2 = self._pseudo_r2_mcfadden(
                y,
                predicted_rate,
                scale=scale,
                aggregate_sample_scores=aggregate_sample_scores,
            )
        elif score_type == "pseudo-r2-Cohen":
            pseudo_r2 = self._pseudo_r2_cohen(
                y, predicted_rate, aggregate_sample_scores=aggregate_sample_scores
            )
        else:
            raise NotImplementedError(f"Score {score_type} not implemented!")
        return pseudo_r2

    def _pseudo_r2_cohen(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Cohen's pseudo-:math:`R^2`.

        Compute the pseudo-:math:`R^2` metric as defined by Cohen et al. (2002). See
        :meth:`nemos.observation_models.Observations.pseudo_r2` for additional information.

        Parameters
        ----------
        y:
            The neural activity. Expected shape: ``(n_time_bins, )``.
        predicted_rate:
            The mean neural activity. Expected shape: ``(n_time_bins, )``

        Returns
        -------
        :
            The pseudo-:math:`R^2` of the model. A value closer to 1 indicates a better model fit,
            whereas a value closer to 0 suggests that the model doesn't improve much over the null model.
        """
        model_dev_t = self.deviance(y, predicted_rate)
        model_deviance = aggregate_sample_scores(model_dev_t)

        null_mu = jnp.ones(y.shape, dtype=jnp.float32) * jnp.mean(y, axis=0)
        null_dev_t = self.deviance(y, null_mu)
        null_deviance = aggregate_sample_scores(null_dev_t)
        return (null_deviance - model_deviance) / null_deviance

    def _pseudo_r2_mcfadden(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        """
        McFadden's pseudo-:math:`R^2`.

        Compute the pseudo-:math:`R^2` metric as defined by McFadden et al. (1979). See
        :meth:`nemos.observation_models.Observations.pseudo_r2` for additional information.

        Parameters
        ----------
        y:
            The neural activity. Expected shape: (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate:
            The mean neural activity. Expected shape: (n_time_bins, ), or (n_time_bins, n_neurons).
        scale:
            The scale parameter of the model.

        Returns
        -------
        :
            The pseudo-:math:`R^2` of the model. A value closer to 1 indicates a better model fit,
            whereas a value closer to 0 suggests that the model doesn't improve much over the null model.
        """
        # ruff: noqa D403
        mean_y = jnp.ones(y.shape) * y.mean(axis=0)
        ll_null = self.log_likelihood(
            y, mean_y, scale=scale, aggregate_sample_scores=aggregate_sample_scores
        )
        ll_model = self.log_likelihood(
            y,
            predicted_rate,
            scale=scale,
            aggregate_sample_scores=aggregate_sample_scores,
        )
        return 1 - ll_model / ll_null


class PoissonObservations(Observations):
    """
    Model observations as Poisson random variables.

    The PoissonObservations is designed to model the observed spike counts based on a Poisson distribution
    with a given rate. It provides methods for computing the negative log-likelihood, generating samples,
    and computing the residual deviance for the given spike count data.

    """

    def __init__(self):
        super().__init__()
        self.scale = 1.0

    @property
    def default_inverse_link_function(self):
        return jnp.exp

    def _negative_log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Compute the Poisson negative log-likelihood.

        This computes the Poisson negative log-likelihood of the predicted rates
        for the observed spike counts up to a constant.

        Parameters
        ----------
        y :
            The target spikes to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons).

        Returns
        -------
        :
            The Poisson negative log-likehood. Shape (1,).

        Notes
        -----
        The formula for the Poisson mean log-likelihood is the following,

        .. math::
        \begin{aligned}
        \text{LL}(\hat{\lambda} | y) &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T}
        [y_{tn} \log(\hat{\lambda}_{tn}) - \hat{\lambda}_{tn} - \log({y_{tn}!})] \\\
        &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T} [y_{tn} \log(\hat{\lambda}_{tn}) -
        \hat{\lambda}_{tn} - \Gamma({y_{tn}+1})] \\\
        &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T} [y_{tn} \log(\hat{\lambda}_{tn}) -
        \hat{\lambda}_{tn}] + \\text{const}
        \end{aligned}

        Because :math:`\Gamma(k+1)=k!`, see `wikipedia <https://en.wikipedia.org/wiki/Gamma_function>` for explanation.

        The :math:`\log({y_{tn}!})` term is not a function of the parameters and can be disregarded
        when computing the loss-function. This is why we incorporated it into the `const` term.
        """
        predicted_rate = jnp.clip(
            predicted_rate, min=jnp.finfo(predicted_rate.dtype).eps
        )
        x = y * jnp.log(predicted_rate)
        # see above for derivation of this.
        return aggregate_sample_scores(predicted_rate - x)

    def log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the Poisson negative log-likelihood.

        This computes the Poisson negative log-likelihood of the predicted rates
        for the observed spike counts up to a constant.

        Parameters
        ----------
        y :
            The target spikes to compare against. Shape ``(n_time_bins, )``, or ``(n_time_bins, n_neurons)``.
        predicted_rate :
            The predicted rate of the current model. Shape ``(n_time_bins, )``, or ``(n_time_bins, n_neurons)``.
        scale :
            The scale parameter of the model.
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The Poisson negative log-likehood. Shape (1,).

        Notes
        -----
        The formula for the Poisson mean log-likelihood is the following,

        .. math::
            \begin{aligned}
            \text{LL}(\hat{\lambda} | y) &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T}
            [y_{tn} \log(\hat{\lambda}_{tn}) - \hat{\lambda}_{tn} - \log({y_{tn}!})] \\\
            &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T} [y_{tn} \log(\hat{\lambda}_{tn}) -
            \hat{\lambda}_{tn} - \Gamma({y_{tn}+1})] \\\
            &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T} [y_{tn} \log(\hat{\lambda}_{tn}) -
            \hat{\lambda}_{tn}] + \text{const}
            \end{aligned}


        Because :math:`\Gamma(k+1)=k!`, see `wikipedia <https://en.wikipedia.org/wiki/Gamma_function>`_ for explanation.

        The :math:`\log({y_{tn}!})` term is not a function of the parameters and can be disregarded
        when computing the loss-function. This is why we incorporated it into the `const` term.
        """
        nll = self._negative_log_likelihood(y, predicted_rate, aggregate_sample_scores)
        return -nll - aggregate_sample_scores(jax.scipy.special.gammaln(y + 1))

    def sample_generator(
        self,
        key: jax.Array,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        """
        Sample from the Poisson distribution.

        This method generates random numbers from a Poisson distribution based on the given
        `predicted_rate`.

        Parameters
        ----------
        key :
            Random key used for the generation of random numbers in JAX.
        predicted_rate :
            Expected rate (lambda) of the Poisson distribution. Shape ``(n_time_bins, )``, or
            ``(n_time_bins, n_neurons)``.
        scale :
            Scale parameter. For Poisson should be equal to 1.

        Returns
        -------
        jnp.ndarray
            Random numbers generated from the Poisson distribution based on the `predicted_rate`.
        """
        return jax.random.poisson(key, predicted_rate)

    def deviance(
        self,
        spike_counts: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        r"""Compute the residual deviance for a Poisson model.

        Parameters
        ----------
        spike_counts:
            The spike counts. Shape ``(n_time_bins, )`` or ``(n_time_bins, n_neurons)`` for population models.
        predicted_rate:
            The predicted firing rates. Shape ``(n_time_bins, )``  or ``(n_time_bins, n_neurons)`` for
            population models.
        scale:
            Scale parameter of the model.

        Returns
        -------
        :
            The residual deviance of the model.

        Notes
        -----
        The deviance is a measure of the goodness of fit of a statistical model.
        For a Poisson model, the residual deviance is computed as:

        .. math::
            \begin{aligned}
                D(y_{tn}, \hat{y}_{tn}) &= 2 \left[ y_{tn} \log\left(\frac{y_{tn}}{\hat{y}_{tn}}\right)
                - (y_{tn} - \hat{y}_{tn}) \right]\\\
                &= 2 \left( \text{LL}\left(y_{tn} | y_{tn}\right) - \text{LL}\left(y_{tn} | \hat{y}_{tn}\right)\right)
            \end{aligned}

        where :math:`y` is the observed data, :math:`\hat{y}` is the predicted data, and :math:`\text{LL}` is
        the model log-likelihood. Lower values of deviance indicate a better fit.
        """
        # this takes care of 0s in the log
        ratio = jnp.clip(
            spike_counts / predicted_rate, jnp.finfo(predicted_rate.dtype).eps, jnp.inf
        )
        deviance = 2 * (spike_counts * jnp.log(ratio) - (spike_counts - predicted_rate))
        return deviance

    def estimate_scale(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        dof_resid: Union[float, jnp.ndarray],
    ) -> Union[float, jnp.ndarray]:
        r"""
        Assign 1 to the scale parameter of the Poisson model.

        For the Poisson exponential family distribution, the scale parameter :math:`\phi` is always 1.
        This property is consistent with the fact that the variance equals the mean in a Poisson distribution.
        As given in the general exponential family expression:

        .. math::
            \text{var}(Y) = \frac{V(\mu)}{a(\phi)},

        for the Poisson family, it simplifies to :math:`\text{var}(Y) = \mu` since :math:`a(\phi) = 1`
        and :math:`V(\mu) = \mu`.

        Parameters
        ----------
        y :
            Observed spike counts.
        predicted_rate :
            The predicted rate values. This is not used in the Poisson model for estimating scale,
            but is retained for compatibility with the abstract method signature.
        dof_resid :
            The DOF of the residuals.
        """
        return jnp.ones_like(jnp.atleast_1d(y[0]))


class GammaObservations(Observations):
    """
    Model observations as Gamma random variables.

    The GammaObservations is designed to model the observed spike counts based on a Gamma distribution
    with a given rate. It provides methods for computing the negative log-likelihood, generating samples,
    and computing the residual deviance for the given spike count data.

    """

    def __init__(
        self,
    ):
        super().__init__()
        self.scale = 1.0

    @property
    def default_inverse_link_function(self):
        return utils.one_over_x

    def _negative_log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Compute the Gamma negative log-likelihood.

        This computes the Gamma negative log-likelihood of the predicted rates
        for the observed neural activity up to a constant.

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The Gamma negative log-likelihood. Shape (1,).

        """
        predicted_rate = jnp.clip(
            predicted_rate, min=jnp.finfo(predicted_rate.dtype).eps
        )
        x = jnp.power(-predicted_rate, -1)
        # see above for derivation of this.
        return -aggregate_sample_scores(y * x + jnp.log(-x))

    def log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the Gamma negative log-likelihood.

        This computes the Gamma negative log-likelihood of the predicted rates
        for the observed neural activity including the normalization constant.

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ) or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ) or (n_time_bins, n_neurons).
        scale :
            The scale parameter of the model.
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The Gamma negative log-likelihood. Shape (1,).

        """
        k = 1 / scale
        norm = (
            (k - 1) * jnp.mean(jnp.log(y))
            + k * jnp.log(k)
            - jax.scipy.special.gammaln(k)
        )
        return aggregate_sample_scores(
            norm - k * self._negative_log_likelihood(y, predicted_rate, lambda x: x)
        )

    def sample_generator(
        self,
        key: jax.Array,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        """
        Sample from the Gamma distribution.

        This method generates random numbers from a Gamma distribution based on the given
        `predicted_rate` and `scale`.

        Parameters
        ----------
        key :
            Random key used for the generation of random numbers in JAX.
        predicted_rate :
            Expected rate (lambda) of the Poisson distribution. Shape (n_time_bins, ), or (n_time_bins, n_neurons)..
        scale:
            The scale parameter for the distribution.

        Returns
        -------
        jnp.ndarray
            Random numbers generated from the Gamma distribution based on the `predicted_rate` and the `scale`.
        """
        return jax.random.gamma(key, predicted_rate / scale) * scale

    def deviance(
        self,
        neural_activity: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        r"""Compute the residual deviance for a Gamma model.

        Parameters
        ----------
        neural_activity:
            The spike coun activity. Shape (n_time_bins, ) or (n_time_bins, n_neurons) for population models.
        predicted_rate:
            The predicted firing rates. Shape (n_time_bins, ) or (n_time_bins, n_neurons) for population models.
        scale:
            Scale parameter of the model.

        Returns
        -------
        :
            The residual deviance of the model.

        Notes
        -----
        The deviance is a measure of the goodness of fit of a statistical model.
        For a Gamma model, the residual deviance is computed as:

        .. math::
            \begin{aligned}
                D(y_{tn}, \hat{y}_{tn}) &=  2 \left[ -\log \frac{ y_{tn}}{\hat{y}_{tn}} +  \frac{y_{tn} -
                \hat{y}_{tn}}{\hat{y}_{tn}}\right]\\\
                &= 2 \left( \text{LL}\left(y_{tn} | y_{tn}\right) - \text{LL}\left(y_{tn} | \hat{y}_{tn}\right) \right)
            \end{aligned}

        where :math:`y` is the observed data, :math:`\hat{y}` is the predicted data, and :math:`\text{LL}` is the model
        log-likelihood. Lower values of deviance indicate a better fit.

        """
        y_mu = jnp.clip(neural_activity / predicted_rate, min=jnp.finfo(float).eps)
        resid_dev = 2 * (
            -jnp.log(y_mu) + (neural_activity - predicted_rate) / predicted_rate
        )
        return resid_dev / scale

    def estimate_scale(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        dof_resid: Union[float, jnp.ndarray],
    ) -> Union[float, jnp.ndarray]:
        r"""
        Estimate the scale of the model based on the GLM residuals.

        For :math:`y \sim \Gamma` the scale is equal to,

        .. math::
            \Phi = \frac{\text{Var(y)}}{V(\mu)}

        with :math:`V(\mu) = \mu^2`.

        Therefore, the scale can be estimated as the ratio of the sample variance to the squared rate.

        Parameters
        ----------
        y :
            Observed neural activity.
        predicted_rate :
            The predicted rate values. This is not used in the Poisson model for estimating scale,
            but is retained for compatibility with the abstract method signature.
        dof_resid :
            The DOF of the residuals.

        Returns
        -------
        :
            The scale parameter. If predicted_rate is ``(n_samples, n_neurons)``, this method will return a
            scale for each neuron.
        """
        predicted_rate = jnp.clip(
            predicted_rate, min=jnp.finfo(predicted_rate.dtype).eps
        )
        resid = jnp.power(y - predicted_rate, 2)
        return (
            jnp.sum(resid * jnp.power(predicted_rate, -2), axis=0) / dof_resid
        )  # pearson residuals


class BernoulliObservations(Observations):
    """
    Model observations as Bernoulli random variables.

    The BernoulliObservations is designed to model an observed binary variable based on a Bernoulli distribution
    with a given success probability. When using a logit link function (i.e. a logistic inverse link function),
    this is equivalent to Logistic Regression. It provides methods for computing the negative log-likelihood,
    generating samples, and computing the residual deviance for the given binary observations.

    """

    def __init__(self):
        super().__init__()
        self.scale = 1.0

    @property
    def default_inverse_link_function(self):
        return jax.lax.logistic

    def _negative_log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Compute the Bernoulli negative log-likelihood.

        This computes the Bernoulli negative log-likelihood of the predicted success probability (predicted rates)
        for the observations up to a constant.

        Parameters
        ----------
        y :
            The target observation to compare against. Shape (n_time_bins, ), or (n_time_bins, n_observations).
        predicted_rate :
            The predicted rate (success probability) of the current model.
            Shape (n_time_bins, ), or (n_time_bins, n_observations).
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The Bernoulli negative log-likelihood. Shape (1,).

        Notes
        -----
        The formula for the Bernoulli mean log-likelihood is the following,

        .. math::
            \text{LL}(p | y) &= \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T}
            [y_{tn} \log(p_{tn}) + (1 - y_{tn}) \log(1 - p_{tn})]

        where :math:`p` is the predicted success probability, given by the inverse link function, and :math:`y` is
        the observed binary variable.
        """
        predicted_rate = jnp.clip(
            predicted_rate,
            min=jnp.finfo(predicted_rate.dtype).eps,
            max=1.0 - jnp.finfo(predicted_rate.dtype).eps,
        )

        x = y * jnp.log(predicted_rate) + (1 - y) * jnp.log1p(-predicted_rate)
        return -aggregate_sample_scores(x)

    def log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the Bernoulli negative log-likelihood.

        This computes the Bernoulli negative log-likelihood of the predicted success probability (predicted rates)
        for the observations up to a constant.

        Parameters
        ----------
        y :
            The target observation to compare against. Shape (n_time_bins, ), or (n_time_bins, n_observations).
        predicted_rate :
            The predicted rate (success probability) of the current model. Shape (n_time_bins, ),
            or (n_time_bins, n_observations).
        scale :
            The scale parameter of the model.
        aggregate_sample_scores :
            Function that aggregates the log-likelihood of each sample.

        Returns
        -------
        :
            The Bernoulli negative log-likelihood. Shape (1,).

        Notes
        -----
        The formula for the Bernoulli mean log-likelihood is the following,

        .. math::
            \text{LL}(p | y) = \frac{1}{T \cdot N} \sum_{n=1}^{N} \sum_{t=1}^{T}
            [y_{tn} \log(p_{tn}) + (1 - y_{tn}) \log(1 - p_{tn})]

        where :math:`p` is the predicted success probability, given by the inverse link function, and :math:`y` is the
        observed binary variable.
        """
        nll = self._negative_log_likelihood(y, predicted_rate, aggregate_sample_scores)
        return -nll

    def sample_generator(
        self,
        key: jax.Array,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        """
        Sample from the Bernoulli distribution.

        This method generates random numbers from a Bernoulli distribution based on the given
        `predicted_rate`.

        Parameters
        ----------
        key :
            Random key used for the generation of random numbers in JAX.
        predicted_rate :
            Expected rate (success probability) of the Poisson distribution. Shape ``(n_time_bins, )``, or
            ``(n_time_bins, n_observations)``.
        scale :
            Scale parameter. For Bernoulli should be equal to 1.

        Returns
        -------
        jnp.ndarray
            Random numbers generated from the Bernoulli distribution based on the `predicted_rate`.
        """
        return jax.random.bernoulli(key, predicted_rate)

    def deviance(
        self,
        observations: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
    ) -> jnp.ndarray:
        r"""Compute the residual deviance for a Bernoulli model.

        Parameters
        ----------
        observations:
            The binary observations. Shape ``(n_time_bins, )`` or ``(n_time_bins, n_observations)`` for population
            models (i.e. multiple observations).
        predicted_rate:
            The predicted rate (success probability). Shape ``(n_time_bins, )``  or ``(n_time_bins, n_observations)``
            for population models (i.e. multiple observations).
        scale:
            Scale parameter of the model. For Bernoulli should be equal to 1.

        Returns
        -------
        :
            The residual deviance of the model.

        Notes
        -----
        The deviance is a measure of the goodness of fit of a statistical model.
        For a Bernoulli model, the residual deviance is computed as:

        .. math::
            \begin{aligned}
                D(y_{tn}, \hat{y}_{tn}) &= 2 \left( \text{LL}\left(y_{tn} | y_{tn}\right) - \text{LL}\left(y_{tn}
                  | \hat{y}_{tn}\right)\right) \\\
                &= 2 \left[ y_{tn} \log\left(\frac{y_{tn}}{\hat{y}_{tn}}\right) + (1 - y_{tn}) \log\left(\frac{1
                  - y_{tn}}{1 - \hat{y}_{tn}}\right) \right]
            \end{aligned}

        where :math:`y` is the observed data, :math:`\hat{y}` is the predicted data, and :math:`\text{LL}` is
        the model log-likelihood. Lower values of deviance indicate a better fit.
        """
        # this takes care of 0s in the log
        ratio1 = jnp.clip(
            observations / predicted_rate, jnp.finfo(predicted_rate.dtype).eps, jnp.inf
        )
        ratio2 = jnp.clip(
            (1 - observations) / (1 - predicted_rate),
            jnp.finfo(predicted_rate.dtype).eps,
            jnp.inf,
        )
        deviance = 2 * (
            observations * jnp.log(ratio1) + (1 - observations) * jnp.log(ratio2)
        )
        return deviance

    def estimate_scale(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        dof_resid: Union[float, jnp.ndarray],
    ) -> Union[float, jnp.ndarray]:
        r"""
        Assign 1 to the scale parameter of the Bernoulli model.

        For the Binomial exponential family distribution (to which the Bernoulli belongs), the scale parameter
        :math:`\phi` is always 1.

        Parameters
        ----------
        y :
            Observed spike counts.
        predicted_rate :
            The predicted rate values (success probabilities). This is not used in the Bernoulli model for estimating
            scale, but is retained for compatibility with the abstract method signature.
        dof_resid :
            The DOF of the residuals.
        """
        return jnp.ones_like(jnp.atleast_1d(y[0]))

    def likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray] = 1.0,
        aggregate_sample_scores: Callable = lambda x: jnp.exp(jnp.mean(jnp.log(x))),
    ):
        r"""Compute the Binomial model likelihood.

        This computes the likelihood of the predicted rates
        for the observed neural activity including the normalization constant.

        Parameters
        ----------
        y :
            The target activity to compare against. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        predicted_rate :
            The predicted rate of the current model. Shape (n_time_bins, ), or (n_time_bins, n_neurons).
        scale :
            The scale parameter of the model
        aggregate_sample_scores :
            Function that aggregates the likelihood of each sample.

        Returns
        -------
        :
            The likelihood. Shape (1,).
        """
        predicted_rate = jnp.clip(
            predicted_rate,
            min=jnp.finfo(predicted_rate.dtype).eps,
            max=1.0 - jnp.finfo(predicted_rate.dtype).eps,
        )
        # convenient formulation that works for specifically for the Bernoulli
        # y can be only 0 or 1 and the likelihood
        # (predicted_rate) ** y * (1 - predicted_rate) ** (1-y)
        # is equal to the computation below, easily shown by plugging y=0,1
        return aggregate_sample_scores(
            y * predicted_rate + (1 - y) * (1 - predicted_rate)
        )


class NegativeBinomialObservations(Observations):
    r"""
    A Negative Binomial model for overdispersed count data using mean-dispersion parameterization.

    This model represents a Negative Binomial distribution [4]_ commonly used to model overdispersed
    count data [5]_ [6]_ (i.e., data where the variance exceeds the mean), which cannot be captured by a
    standard Poisson model. The distribution is parameterized by the predicted mean rate
    (:math:`\mu`) and a fixed dispersion parameter (:math:`\phi`) or `scale` of the model.

    **Important:** the scale parameter must be estimated from the data for accurately capturing the
    dispersion. In the context of NeMoS GLM, estimation can be achieved by cross-validating the
    ``scale`` parameter. One may use scikit-learn :ref:`GridSearchCV <sklearn-how-to>` for example.

    The variance of the Negative Binomial distribution under this parameterization is:

    .. math::

        \mathrm{Var}(Y) = \mu + \phi \mu^2

    where :math:`\mu` is the predicted mean, and :math:`\phi` is the dispersion parameter. This
    formulation corresponds to the Negative Binomial as a Gamma–Poisson mixture.

    The scale parameter :math:`\phi` is related to the canonical Negative Binomial
    shape parameter `r` as:

    .. math::

        r = \frac{1}{\phi}

    As :math:`\phi \to 0` (equivalently, :math:`r \to \infty`), the distribution approaches a
    Poisson distribution. This makes the model flexible for handling both equidispersed
    (Poisson-like) and overdispersed data.

    Parameters
    ----------
    scale :
        The dispersion parameter :math:`\phi`. Lower values correspond to lower overdispersion, and as
        :math:`\phi \to 0`, the model behaves like a Poisson. The shape parameter of
        the Negative Binomial is given by `r = 1 / scale`.

    References
    ----------
    .. [4] https://en.wikipedia.org/wiki/Negative_binomial_distribution

    .. [5] Pillow, Jonathan, and James Scott. "Fully Bayesian inference for neural models
        with negative-binomial spiking." Advances in neural information processing systems 25 (2012).

    .. [6] Wei, Ganchao, et al. "Calibrating Bayesian decoders of neural spiking activity."
        Journal of Neuroscience 44.18 (2024).
    """

    def __init__(
        self,
        scale=1.0,
    ):
        super().__init__()
        self.scale = scale

    @property
    def default_inverse_link_function(self):
        return jnp.exp

    def _negative_log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        aggregate_sample_scores: Callable = jnp.mean,
    ) -> jnp.ndarray:
        r"""Compute the Negative Binomial negative log-likelihood.

        This computes the Negative Binomial negative log-likelihood of the
        predicted mean rate for the observed counts.

        Parameters
        ----------
        y :
            Observed count data. Shape ``(n_time_bins,)`` or ``(n_time_bins, n_observations)``.
        predicted_rate :
            The predicted mean of the Negative Binomial distribution. Shape
            ``(n_time_bins,)`` or ``(n_time_bins, n_observations)``.
        aggregate_sample_scores :
            Function that aggregates the log-likelihood across samples (e.g., ``jnp.mean`` or
            ``jnp.sum``).

        Returns
        -------
        :
            The Negative Binomial negative log-likelihood. Shape ``(1,)``.

        Notes
        -----
        The Negative Binomial distribution models overdispersed count data.
        The likelihood assumes the mean-parameterized form with dispersion `r = 1 / scale`.

        The log-likelihood is computed (up to a constant) using:

        .. math::
            \log p(y | \mu, r) = \log \Gamma(y + r) - \log \Gamma(r) - \log y!
            + r \log\left(\frac{r}{r + \mu}\right)
            + y \log\left(\frac{\mu}{r + \mu}\right)

        """
        if self.scale is None:
            self.estimate_scale(y, predicted_rate, aggregate_sample_scores)
        predicted_rate = jnp.clip(
            predicted_rate, min=jnp.finfo(predicted_rate.dtype).eps
        )
        factor = 1 / (self.scale * predicted_rate + 1)
        return -aggregate_sample_scores(
            y * jnp.log(1 - factor) + jnp.log(factor) / self.scale
        )

    def log_likelihood(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray, None] = None,
        aggregate_sample_scores: Callable = jnp.mean,
    ):
        r"""Compute the Negative Binomial log-likelihood.

        This computes the Negative Binomial log-likelihood of the predicted mean
        rate for the observed counts.

        Parameters
        ----------
        y :
            Observed count data. Shape ``(n_time_bins,)`` or ``(n_time_bins, n_observations)``.
        predicted_rate :
            The predicted mean of the Negative Binomial distribution. Shape ``(n_time_bins,)`` or
            ``(n_time_bins, n_observations)``.
        scale :
            The scale (dispersion) parameter of the distribution. It is related to the shape ``r`` as ``r = 1 / scale``.
            Default is the scale provided at initialization ``self.scale``.
        aggregate_sample_scores :
            Function that aggregates the log-likelihood across samples (e.g., jnp.mean or jnp.sum).

        Returns
        -------
        :
            The log-likelihood of the Negative Binomial model. Shape ``(1,)``.
        """
        scale = self.scale if scale is None else scale
        ll_unnormalized = -self._negative_log_likelihood(
            y, predicted_rate, aggregate_sample_scores
        )
        norm = aggregate_sample_scores(
            jax.scipy.special.gammaln(y + 1 / scale)
            - jax.scipy.special.gammaln(y + 1)
            - jax.scipy.special.gammaln(1 / scale)
        )
        return ll_unnormalized + norm

    def sample_generator(
        self,
        key: jax.Array,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray, None] = None,
    ) -> jnp.ndarray:
        r"""
        Sample from the Negative Binomial distribution.

        This method generates random count data from the Negative Binomial distribution based on the predicted rate
        and scale (dispersion).

        Parameters
        ----------
        key :
            Random key used for number generation in JAX.
        predicted_rate :
            The predicted mean of the Negative Binomial distribution. Shape ``(n_time_bins,)`` or
            ``(n_time_bins, n_observations)``.
        scale :
            Dispersion parameter of the distribution. Smaller values imply higher variance.

        Returns
        -------
        :
            Samples drawn from the Negative Binomial distribution. Same shape as ``predicted_rate``.

        Notes
        -----
        This method uses the Gamma--Poisson mixture representation of the Negative Binomial distribution:

        .. math::

            Y \sim \text{Poisson}(\lambda), \quad \lambda \sim \text{Gamma}(r, \beta)

        where :math:`r = 1 / \phi` and :math:`\beta = r / \mu`.

        For more information, see the `Negative Binomial distribution on Wikipedia
        <https://en.wikipedia.org/wiki/Negative_binomial_distribution#Gamma%E2%80%93Poisson_mixture>`_.
        """
        scale = self.scale if scale is None else scale
        r = 1.0 / scale
        gamma_key, poisson_key = jax.random.split(key)

        # Gamma with shape=r, rate=r/mu → scale=mu/r
        gamma_sample = jax.random.gamma(gamma_key, r, shape=predicted_rate.shape) * (
            predicted_rate / r
        )

        return jax.random.poisson(poisson_key, gamma_sample)

    def deviance(
        self,
        observations: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        scale: Union[float, jnp.ndarray, None] = None,
    ) -> jnp.ndarray:
        r"""Compute the residual deviance for a Negative Binomial model.

        The deviance measures how well a statistical model fits the data by
        quantifying the difference between the observed values and the values
        predicted by the model. Lower values of deviance indicate a better fit.

        Parameters
        ----------
        observations:
            Observed count data. Shape ``(n_time_bins,)`` or ``(n_time_bins, n_observations).``
        predicted_rate:
            Predicted mean count of the Negative Binomial distribution. Shape matches `observations`.
        scale:
            Dispersion parameter of the distribution.

        Returns
        -------
        :
            The residual deviance of the model. Shape matches ``observations``.

        Notes
        -----
        The deviance is a measure of the goodness of fit of a statistical model.
        For a Negative Binomial model, the residual deviance is computed as:

        .. math::
            \begin{aligned}
                D(y_{tn}, \hat{y}_{tn}) &= 2 \left( \text{LL}\left(y_{tn} | y_{tn}\right) - \text{LL}\left(y_{tn}
                  | \hat{y}_{tn}\right)\right) \\\
                &= 2 \left[ y_{tn} \log\left(\frac{y_{tn}}{\hat{y}_{tn}}\right) + (1 - y_{tn}) \log\left(\frac{1
                  - y_{tn}}{1 - \hat{y}_{tn}}\right) \right]
            \end{aligned}

        where :math:`y` is the observed data, :math:`\hat{y}` is the predicted data, and :math:`\text{LL}` is
        the model log-likelihood.
        """
        scale = self.scale if scale is None else scale
        factor = (predicted_rate * scale + 1) / (observations * scale + 1)
        y_mu = observations / predicted_rate
        y_mu = jnp.clip(y_mu, min=jnp.finfo(predicted_rate.dtype).eps)
        term1 = observations * jnp.log(y_mu * factor)
        term2 = jnp.log(factor) / scale
        return 2 * (term1 + term2)

    def estimate_scale(
        self,
        y: jnp.ndarray,
        predicted_rate: jnp.ndarray,
        dof_resid: Union[float, jnp.ndarray],
    ) -> Union[float, jnp.ndarray]:
        r"""
        Return the scale parameter of the distribution.

        The ``scale`` parameter of the Negative Binomial distribution is set
        at initialization and affect the likelihood landscape. This implies
        that the ``scale`` parameter cannot be estimated post-hoc without
        re-fitting a model.

        Note that the arguments of this method are not used but are kept for
        API consistency—i.e., all ``Observations.estimate_scale`` methods
        have the same signature.

        Parameters
        ----------
        y :
            Observed spike counts.
        predicted_rate :
            The predicted mean of the distribution.
        dof_resid :
            The DOF of the residuals.

        Notes
        -----
        NeMoS currently does not support joint estimation of scale and mean for the negative binomial.
        For alternatives, see the R package MASS
        `glm.nb <https://www.rdocumentation.org/packages/MASS/versions/7.3-65/topics/glm.nb>`_
        for more details.
        """
        return jnp.array(self.scale)


def check_observation_model(observation_model, force_checks=False):
    r"""
    Check the attributes of an observation model for compliance.

    This function ensures that the observation model has the required attributes and that each
    attribute is a callable function. Additionally, it checks if these functions return
    jax.numpy.ndarray objects, and in the case of 'inverse_link_function', whether it is
    differentiable.

    Parameters
    ----------
    observation_model : object
        An instance of an observation model that should have specific attributes.
    force_checks:
        If true, always checks. This is intended for testing purposes, to make sure
        that the check passes for native nemos observation models.

    Raises
    ------
    AttributeError
        If the `observation_model` does not have one of the required attributes.

    TypeError
        If an attribute is not a callable function.
    TypeError
        If a function does not return a jax.numpy.ndarray.
    TypeError
        If 'inverse_link_function' is not differentiable.

    Examples
    --------
    >>> import jax
    >>> import jax.numpy as jnp
    >>> class MyObservationModel:
    ...     @property
    ...     def default_inverse_link_function(self):
    ...         return jax.scipy.special.expit
    ...     def _negative_log_likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
    ...         return -aggregate_sample_scores(y * jax.scipy.special.logit(rate) + \
    ...                                         (1 - y) * jax.scipy.special.logit(1 - rate))
    ...     def pseudo_r2(self, y, rate, aggregate_sample_scores=jnp.mean):
    ...         return 1 - (self._negative_log_likelihood(y, rate, aggregate_sample_scores) /
    ...                     jnp.sum((y - y.mean()) ** 2))
    ...     def sample_generator(self, key, rate, scale=1.):
    ...         return jax.random.bernoulli(key, rate)
    ...     def estimate_scale(self, y, predicted_rate, dof_resid):
    ...         return jnp.array(1.)
    ...     def log_likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
    ...         return -self._negative_log_likelihood(y, rate, aggregate_sample_scores)
    ...     def likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
    ...         return jnp.exp(self.log_likelihood(y, rate, aggregate_sample_scores))
    ...     def deviance(self, y, rate, scale=1.0):
    ...         identity = lambda x: x
    ...         return 2 * (
    ...                 self.log_likelihood(y, rate, identity) -
    ...                 self.log_likelihood(y.mean()*jnp.ones_like(y), rate, identity))
    >>> model = MyObservationModel()
    >>> check_observation_model(model)  # Should pass without error if the model is correctly implemented.
    """
    # Define the checks to be made on each attribute

    is_nemos = isinstance(
        observation_model,
        (
            PoissonObservations,
            GammaObservations,
            BernoulliObservations,
            NegativeBinomialObservations,
        ),
    )

    checks = {}
    if not is_nemos or force_checks:
        checks.update(
            {
                "_negative_log_likelihood": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                    ],
                    "test_scalar_func": True,
                },
                "pseudo_r2": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                    ],
                    "test_scalar_func": True,
                },
                "log_likelihood": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                    ],
                    "test_scalar_func": True,
                },
                "likelihood": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                    ],
                    "test_scalar_func": True,
                },
                "deviance": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                    ],
                    "test_scalar_func": False,
                },
                "estimate_scale": {
                    "input": [
                        0.5 * jnp.array([1.0, 1.0, 1.0]),
                        jnp.array([1.0, 1.0, 1.0]),
                        1,
                    ],
                    "test_scalar_func": False,
                },
                "sample_generator": {
                    "input": [jax.random.key(123), 0.5 * jnp.array([1.0, 1.0, 1.0]), 1],
                    "test_preserve_shape": True,
                },
                "default_inverse_link_function": {
                    "input": [jnp.array([1.0, 1.0, 1.0])],
                    "test_preserve_shape": False,
                },
            }
        )

    # Perform checks for each attribute
    for attr_name, check_info in checks.items():

        # check if the observation model has the attribute
        utils.assert_has_attribute(observation_model, attr_name)

        # check if the attribute is a callable
        func = getattr(observation_model, attr_name)
        utils.assert_is_callable(func, attr_name)

        # check that the callable returns an array
        utils.assert_returns_ndarray(func, check_info["input"], attr_name)

        if check_info.get("test_differentiable"):
            utils.assert_differentiable(func, attr_name)

        if "test_preserve_shape" in check_info:
            index = int(check_info["test_preserve_shape"])
            utils.assert_preserve_shape(
                func, check_info["input"], attr_name, input_index=index
            )

        if check_info.get("test_scalar_func"):
            utils.assert_scalar_func(func, check_info["input"], attr_name)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from absl.testing import absltest

import jax
import jax.numpy as jnp
from jax.test_util import check_grads

from nemos.third_party.jaxopt.jaxopt import linear_solve
from nemos.third_party.jaxopt.jaxopt import IterativeRefinement
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class IterativeRefinementTest(test_util.JaxoptTestCase):

  def test_simple_system(self):
    onp.random.seed(0)
    n = 20
    A = onp.random.rand(n, n)
    b = onp.random.randn(n)

    low_acc = 1e-1
    high_acc = 1e-5

    # Heavily regularized low acuracy solver.
    inner_solver = partial(linear_solve.solve_gmres, tol=low_acc, ridge=1e-3)

    solver = IterativeRefinement(solve=inner_solver, tol=high_acc, maxiter=10)
    x, state = solver.run(None, A, b)
    self.assertLess(state.error, high_acc)

    x_approx = inner_solver(lambda x: jnp.dot(A, x), b)
    error_inner_solver = solver.l2_optimality_error(x_approx, A, b)
    # High accuracy solution obtained from low accuracy solver.
    self.assertLess(state.error, error_inner_solver)

  def test_ill_posed_problem(self):
    onp.random.seed(0)
    n = 10
    e = 5

    # duplicated rows.
    A = onp.random.rand(e, n)
    A = jnp.concatenate([A, A], axis=0)
    b = onp.random.randn(e)
    b = jnp.concatenate([b, b], axis=0)

    low_acc = 1e-1
    high_acc = 1e-3

    # Heavily regularized low acuracy solver.
    inner_solver = partial(linear_solve.solve_gmres, tol=low_acc, ridge=5e-2)

    solver = IterativeRefinement(solve=inner_solver, tol=high_acc, maxiter=30)
    x, state = solver.run(init_params=None, A=A, b=b)
    self.assertLess(state.error, high_acc)

    x_approx = inner_solver(lambda x: jnp.dot(A, x), b)
    error_inner_solver = solver.l2_optimality_error(x_approx, A, b)
    # High accuracy solution obtained from low accuracy solver.
    self.assertLess(state.error, error_inner_solver)

  def test_perturbed_system(self):
    onp.random.seed(0)
    n = 20

    A = onp.random.rand(n, n)  # invertible matrix (with high probability).

    noise = onp.random.randn(n, n)
    sigma = 0.05
    A_bar = A + sigma * noise  # perturbed system.

    expected = onp.random.randn(n)
    b = A @ expected  # unperturbed target.

    high_acc = 1e-3
    solver = IterativeRefinement(matvec_A=None, matvec_A_bar=jnp.dot,
                                 tol=high_acc, maxiter=100)
    x, state = solver.run(init_params=None, A=A, b=b, A_bar=A_bar)
    self.assertLess(state.error, high_acc)
    self.assertArraysAllClose(x, expected, rtol=5e-2)

  def test_implicit_diff(self):
    onp.random.seed(17)
    n = 20
    A = onp.random.rand(n, n)
    b = onp.random.randn(n)

    low_acc = 1e-1
    high_acc = 1e-5

    # Heavily regularized low acuracy solver.
    inner_solver = partial(linear_solve.solve_gmres, tol=low_acc, ridge=1e-3)
    solver = IterativeRefinement(solve=inner_solver, tol=high_acc, maxiter=10)

    def solve_run(A, b):
      x, state = solver.run(init_params=None, A=A, b=b)
      return x

    check_grads(solve_run, args=(A, b), order=1, modes=['rev'], eps=1e-3)

  def test_warm_start(self):
    onp.random.seed(0)
    n = 20
    A = onp.random.rand(n, n)
    b = onp.random.randn(n)

    init_x = onp.random.randn(n)

    high_acc = 1e-5

    solver = IterativeRefinement(tol=high_acc, maxiter=10)
    x, state = solver.run(init_x, A, b)
    self.assertLess(state.error, high_acc)


if __name__ == "__main__":
  jax.config.update("jax_enable_x64", False)  # low precision environment.
  absltest.main()



# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
import copy
import math
import warnings
from copy import deepcopy
from functools import wraps
from typing import Callable, Generator, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..base_class import Base
from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ..utils import row_wise_kron
from ..validation import check_fraction_valid_samples
from ._basis_mixin import BasisMixin, BasisTransformerMixin, CompositeBasisMixin
from ._check_basis import (
    _check_input_dimensionality,
    _check_transform_input,
    _check_zero_samples,
)
from ._composition_utils import (
    _check_unique_shapes,
    _have_unique_shapes,
    add_docstring,
    get_input_shape,
    infer_input_dimensionality,
    is_basis_like,
    multiply_basis_by_integer,
    promote_to_transformer,
    raise_basis_to_power,
    set_input_shape,
)


def check_transform_input(func: Callable) -> Callable:
    """Check input before calling basis.

    This decorator allows to raise an exception that is more readable
    when the wrong number of input is provided to evaluate.
    """

    @wraps(func)
    def wrapper(self: Basis, *xi: ArrayLike, **kwargs) -> NDArray:
        xi = self._check_transform_input(*xi)
        return func(self, *xi, **kwargs)  # Call the basis

    return wrapper


def check_one_dimensional(func: Callable) -> Callable:
    """Check if the input is one-dimensional."""

    @wraps(func)
    def wrapper(self: Basis, *xi: NDArray, **kwargs):
        if any(x.ndim != 1 for x in xi):
            raise ValueError("Input sample must be one dimensional!")
        return func(self, *xi, **kwargs)

    return wrapper


def min_max_rescale_samples(
    sample_pts: NDArray,
    bounds: Optional[Tuple[float, float]] = None,
) -> Tuple[NDArray, NDArray]:
    """Rescale samples to [0,1].

    Parameters
    ----------
    sample_pts:
        The original samples.
    bounds:
        Sample bounds. `bounds[0]` and `bounds[1]` are mapped to 0 and 1, respectively.
        Default are `min(sample_pts), max(sample_pts)`.

    Warns
    -----
    UserWarning
        If more than 90% of the sample points contain NaNs or Infs.
    """
    sample_pts = sample_pts.astype(float)
    # if not normalize all array
    vmin = np.nanmin(sample_pts, axis=0) if bounds is None else bounds[0]
    vmax = np.nanmax(sample_pts, axis=0) if bounds is None else bounds[1]
    sample_pts[(sample_pts < vmin) | (sample_pts > vmax)] = np.nan
    sample_pts -= vmin

    scaling = np.asarray(vmax - vmin)
    # do not normalize if samples contain a single value (in which case vmax=vmin)
    scaling[scaling == 0] = 1.0
    sample_pts /= scaling

    check_fraction_valid_samples(
        sample_pts,
        err_msg="All the samples lie outside the [vmin, vmax] range.",
        warn_msg="More than 90% of the samples lie outside the [vmin, vmax] range.",
    )

    return sample_pts, scaling


def get_equi_spaced_samples(
    *n_samples,
    bounds: Optional[tuple[float, float] | tuple[tuple[float, float]]] = None,
) -> Generator[NDArray]:
    """Get equi-spaced samples for all the input dimensions.

    This will be used to evaluate the basis on a grid of
    points derived by the samples.

    Parameters
    ----------
    n_samples[0],...,n_samples[n]
        The number of samples in each axis of the grid.
    bounds:
        The bounds for the linspace, if provided.

    Returns
    -------
    :
        A generator yielding numpy arrays of linspaces from 0 to 1 of sizes specified by ``n_samples``.
    """
    # handling of defaults when evaluating on a grid
    # (i.e. when we cannot use max and min of samples)
    if bounds is None:
        mn, mx = 0, 1
    elif all(isinstance(b, tuple) and len(b) == 2 for b in bounds):
        return (np.linspace(*b, samp) for b, samp in zip(bounds, n_samples))
    else:
        mn, mx = bounds
    return (np.linspace(mn, mx, samp) for samp in n_samples)


class Basis(Base, abc.ABC, BasisTransformerMixin):
    """
    Abstract base class for defining basis functions for feature transformation.

    Basis functions are mathematical constructs that can represent data in alternative,
    often more compact or interpretable forms. This class provides a template for such
    transformations, with specific implementations defining the actual behavior.

    Raises
    ------
    ValueError:
        If ``kwargs`` include parameters not recognized or do not have
        default values in ``create_convolutional_predictor``.
    ValueError:
        If ``axis`` different from 0 is provided as a keyword argument (samples must always be in the first axis).
    """

    def __init__(
        self,
    ) -> None:
        self._n_input_dimensionality = getattr(self, "_n_input_dimensionality", 0)

        # specified only after inputs/input shapes are provided
        self._input_shape_product = getattr(self, "_input_shape_product", None)

    @property
    def n_basis_funcs(self):
        """Number of basis functions."""
        return self._n_basis_funcs

    @n_basis_funcs.setter
    def n_basis_funcs(self, value):
        orig_n_basis = copy.deepcopy(getattr(self, "_n_basis_funcs", None))
        self._n_basis_funcs = value
        try:
            self._check_n_basis_min()
        except ValueError as e:
            self._n_basis_funcs = orig_n_basis
            raise e

    @check_transform_input
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Apply the basis transformation to the input data.

        This method is designed to be a high-level interface for transforming input
        data using the basis functions defined by the subclass. Depending on the basis'
        mode ('Eval' or 'Conv'), it either evaluates the basis functions at the sample
        points or performs a convolution operation between the input data and the
        basis functions.

        Parameters
        ----------
        *xi :
            Input data arrays to be transformed. The shape and content requirements
            depend on the subclass and mode of operation ('Eval' or 'Conv').

        Returns
        -------
        :
            Transformed features. In 'Eval' mode, it corresponds to the basis functions
            evaluated at the input samples. In 'Conv' mode, it consists of convolved
            input samples with the basis functions. The output shape varies based on
            the subclass and mode.

        Notes
        -----
        Subclasses should implement how to handle the transformation specific to their
        basis function types and operation modes.
        """
        self.setup_basis(*xi)
        return self._compute_features(*xi)

    @abc.abstractmethod
    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """Convolve or evaluate the basis.

        This method is intended to be equivalent to the sklearn transformer ``transform`` method.
        As the latter, it computes the transformation assuming that all the states are already
        pre-computed by ``_fit_basis``, a method corresponding to ``fit``.

        The method differs from  transformer's ``transform`` for the structure of the input that it accepts.
        In particular, ``_compute_features`` accepts a number of different time series, one per 1D basis component,
        while ``transform`` requires all inputs to be concatenated in a single array.
        """
        pass

    @abc.abstractmethod
    def setup_basis(self, *xi: ArrayLike) -> FeatureMatrix:
        """Pre-compute all basis state variables.

        This method is intended to be equivalent to the sklearn transformer ``fit`` method.
        As the latter, it computes all the state attributes, and store it with the convention
        that the attribute name **must** end with "_", for example ``self.kernel_``,
        ``self._input_shape_``.

        The method differs from  transformer's ``fit`` for the structure of the input that it accepts.
        In particular, ``_fit_basis`` accepts a number of different time series, one per 1D basis component,
        while ``fit`` requires all inputs to be concatenated in a single array.
        """
        pass

    @abc.abstractmethod
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Set the expected input shape for the basis object.

        This method configures the shape of the input data that the basis object expects.
        ``xi`` can be specified as an integer, a tuple of integers, or derived
        from an array. The method also calculates the total number of input
        features and output features based on the number of basis functions.

        """
        pass

    @abc.abstractmethod
    def evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Abstract method to evaluate the basis functions at given points.

        This method must be implemented by subclasses to define the specific behavior
        of the basis transformation. The implementation depends on the type of basis
        (e.g., spline, raised cosine), and it should evaluate the basis functions at
        the specified points in the domain.

        Parameters
        ----------
        *xi :
            Variable number of arguments, each representing an array of points at which
            to evaluate the basis functions. The dimensions and requirements of these
            inputs vary depending on the specific basis implementation.

        Returns
        -------
        :
            An array containing the evaluated values of the basis functions at the input
            points. The shape and structure of this array are specific to the subclass
            implementation.
        """
        pass

    def _get_samples(self, *n_samples: int) -> Generator[NDArray]:
        """Get equi-spaced samples for all the input dimensions.

        This will be used to evaluate the basis on a grid of
        points derived by the samples.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of samples in each axis of the grid.

        Returns
        -------
        :
            A generator yielding numpy arrays of linspaces from 0 to 1 of sizes specified by ``n_samples``.
        """
        # handling of defaults when evaluating on a grid
        # (i.e. when we cannot use max and min of samples)
        bounds = getattr(self, "bounds", None)
        return get_equi_spaced_samples(*n_samples, bounds=bounds)

    def _check_transform_input(
        self, *xi: ArrayLike
    ) -> Tuple[Union[NDArray, Tsd, TsdFrame]]:
        """Check transform input.

        Parameters
        ----------
        xi[0],...,xi[n] :
            The input samples, each  with shape (number of samples, ).

        Raises
        ------
        ValueError
            - If the time point number is inconsistent between inputs.
            - If the number of inputs doesn't match what the Basis object requires.
            - At least one of the samples is empty.

        """
        # standard checks and transform to array
        inp = _check_transform_input(self, *xi)
        input_idx = 0
        for b in self:
            # check if exact shape matching for multiplicative bases
            if isinstance(b, MultiplicativeBasis):
                n_input = infer_input_dimensionality(b)
                b_input = inp[input_idx : input_idx + n_input]
                _check_unique_shapes(b_input, basis=b)
                input_idx += n_input
        return inp

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
           The number of samples.

        Returns
        -------
        X :
           Array of shape ``(n_samples,)`` containing the equi-spaced sample
           points where we've evaluated the basis.
        basis_funcs :
           Evaluated exponentially decaying basis functions, numerically
           orthogonalized, shape ``(n_samples, n_basis_funcs)``
        """
        _check_input_dimensionality(self, n_samples)

        _check_zero_samples(
            n_samples,
            err_message="All sample counts provided must be greater than zero.",
        )

        # get the samples (can be re-implemented, by providing a _get_samples)
        bounds = getattr(self, "bounds", None)
        get_samples = getattr(
            self, "_get_samples", lambda *x: get_equi_spaced_samples(*x, bounds=bounds)
        )
        sample_tuple = get_samples(*n_samples)
        Xs = np.meshgrid(*sample_tuple, indexing="ij")

        # evaluates the basis on a flat NDArray and reshape to match meshgrid output
        Y = self.evaluate(*(grid_axis.flatten() for grid_axis in Xs)).reshape(
            (*n_samples, self.n_basis_funcs)
        )

        return *Xs, Y

    @promote_to_transformer
    def __add__(self, other: BasisMixin) -> AdditiveBasis:
        """
        Add two Basis objects together.

        Parameters
        ----------
        other
            The other Basis object to add.

        Returns
        -------
        : AdditiveBasis
            The resulting Basis object.
        """
        return AdditiveBasis(self, other)

    @promote_to_transformer
    def __rmul__(self, other: BasisMixin | int):
        """Right multiplication operator for basis."""
        return self.__mul__(other)

    @promote_to_transformer
    def __mul__(self, other: BasisMixin | int) -> Basis:
        """
        Multiply two Basis objects together.

        Parameters
        ----------
        other
            The other Basis object to multiply.

        Returns
        -------
        :
            The resulting Basis object.
        """
        if isinstance(other, int):
            return multiply_basis_by_integer(self, other)

        if not is_basis_like(other):
            raise TypeError(
                "Basis multiplicative factor should be a Basis object or a positive integer!"
            )
        return MultiplicativeBasis(self, other)

    @promote_to_transformer
    def __pow__(self, exponent: int) -> BasisMixin:
        """Exponentiation of a Basis object.

        Define the power of a basis by repeatedly applying the method __multiply__.
        The exponent must be a positive integer.

        Parameters
        ----------
        exponent :
            Positive integer exponent

        Returns
        -------
        :
            The product of the basis with itself "exponent" times. Equivalent to ``self * self * ... * self``.

        Raises
        ------
        TypeError
            If the provided exponent is not an integer.
        ValueError
            If the integer is zero or negative.
        """
        return raise_basis_to_power(self, exponent)

    def __len__(self):
        """Return the number of additive basis."""
        return 1


class AdditiveBasis(CompositeBasisMixin, Basis):
    """
    Class representing the addition of two Basis objects.

    Parameters
    ----------
    basis1 :
        First basis object to add.
    basis2 :
        Second basis object to add.

    Examples
    --------
    >>> # Generate sample data
    >>> import numpy as np
    >>> import nemos as nmo
    >>> X = np.random.normal(size=(30, 2))

    >>> # define two basis objects and add them
    >>> basis_1 = nmo.basis.BSplineEval(10)
    >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
    >>> additive_basis = basis_1 + basis_2
    >>> additive_basis
    '(BSplineEval + RaisedCosineLinearEval)': AdditiveBasis(
        ...
    )
    >>> # can add another basis to the AdditiveBasis object
    >>> X = np.random.normal(size=(30, 3))
    >>> basis_3 = nmo.basis.RaisedCosineLogEval(100)
    >>> additive_basis_2 = additive_basis + basis_3
    >>> additive_basis_2
    '((BSplineEval + RaisedCosineLinearEval) + RaisedCosineLogEval)': AdditiveBasis(
        ...
    )
    """

    def __init__(
        self, basis1: BasisMixin, basis2: BasisMixin, label: Optional[str] = None
    ) -> None:
        CompositeBasisMixin.__init__(self, basis1, basis2, label=label)
        Basis.__init__(self)

    def _generate_label(self) -> str:
        return "(" + self.basis1.label + " + " + self.basis2.label + ")"

    @property
    def n_basis_funcs(self):
        """Compute the n-basis function runtime.

        This plays well with cross-validation where the number of basis function of the
        underlying bases can be changed. It must be read-only since the number of basis
        is determined by the two basis elements and the type of composition.
        """
        return self.basis1.n_basis_funcs + self.basis2.n_basis_funcs

    @property
    def n_output_features(self):
        """Return the number of output features."""
        out1 = getattr(self.basis1, "n_output_features", None)
        out2 = getattr(self.basis2, "n_output_features", None)
        if out1 is None or out2 is None:
            return None
        return out1 + out2

    def set_input_shape(self, *xi: int | tuple[int, ...] | NDArray) -> Basis:
        """
        Set the expected input shape for the basis object.

        This method sets the input shape for each component basis in the basis.
        One ``xi`` must be provided for each basis component, specified as an integer,
        a tuple of integers, or an array. The method calculates and stores the total number of output features
        based on the number of basis functions in each component and the provided input shapes.

        Parameters
        ----------
        *xi :
            The input shape specifications. For every k,``xi[k]`` can be:
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided, and it contains non-integer elements.
            If not enough inputs are provided.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.

        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define an additive basis
        >>> basis_1 = nmo.basis.BSplineEval(5)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(6)
        >>> basis_3 = nmo.basis.RaisedCosineLinearEval(7)
        >>> additive_basis = basis_1 + basis_2 + basis_3

        Specify the input shape using all 3 allowed ways: integer, tuple, array
        >>> _ = additive_basis.set_input_shape(1, (2, 3), np.ones((10, 4, 5)))

        Expected output features are:
        (5 bases * 1 input) + (6 bases * 6 inputs) + (7 bases * 20 inputs) = 181
        >>> additive_basis.n_output_features
        181

        """
        # ruff: noqa: D205, D400
        return super().set_input_shape(*xi)

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    @check_one_dimensional
    def evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Evaluate the basis at the sample points.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs)

        Notes
        -----
            Each additive component can process inputs of different shapes, as long as the
            sample axis matches. Input of mis-matched shape cannot be concatenated, therefore
            we enforce 1-dimensional input only for evaluate.

        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo
        >>> x, y = np.random.normal(size=(2, 30))

        >>> # define two basis objects and add them
        >>> basis_1 = nmo.basis.BSplineEval(10)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
        >>> additive_basis = basis_1 + basis_2

        >>> # call the basis.
        >>> out = additive_basis.evaluate(x, y)

        """
        X = np.hstack(
            (
                self.basis1.evaluate(*xi[: self.basis1._n_input_dimensionality]),
                self.basis2.evaluate(*xi[self.basis1._n_input_dimensionality :]),
            )
        )
        return X

    @add_docstring("compute_features", Basis)
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_add = basis1 + basis2
        >>> X_multi = basis_add.compute_features(np.random.randn(20), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 17 = 5 + 2*6
        (20, 17)

        """
        # ruff: noqa: D205, D400
        return super().compute_features(*xi)

    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Compute features for added bases and concatenate.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The features, shape (n_samples, n_basis_funcs)

        """
        # the numpy conversion is important, there is some in-place
        # array modification in basis.
        hstack_pynapple = support_pynapple(conv_type="numpy")(np.hstack)
        comp_feature_1 = getattr(
            self.basis1, "_compute_features", self.basis1.compute_features
        )
        comp_feature_2 = getattr(
            self.basis2, "_compute_features", self.basis2.compute_features
        )
        X = hstack_pynapple(
            (
                comp_feature_1(*xi[: self.basis1._n_input_dimensionality]),
                comp_feature_2(*xi[self.basis1._n_input_dimensionality :]),
            ),
        )
        return X

    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Decompose an array along a specified axis into sub-arrays based on the basis components.

        This function takes an array (e.g., a design matrix or model coefficients) and splits it along
        a designated axis. Each split corresponds to a different additive component of the basis,
        preserving all dimensions except the specified axis.

        **How It Works:**

        Suppose the basis is made up of **m components**, each with :math:`b_i` basis functions and :math:`n_i` inputs.
        The total number of features, :math:`N`, is calculated as:

        .. math::
            N = b_1 \cdot n_1 + b_2 \cdot n_2 + \ldots + b_m \cdot n_m

        This method splits any axis of length :math:`N` into sub-arrays, one for each basis component.

        The sub-array for the i-th basis component is reshaped into dimensions
        :math:`(n_i, b_i)`.

        For example, if the array shape is :math:`(1, 2, N, 4, 5)`, then each split sub-array will have shape:

        .. math::
            (1, 2, n_i, b_i, 4, 5)

        where:

        - :math:`n_i` represents the number of inputs associated with the i-th component,
        - :math:`b_i` represents the number of basis functions in that component.

        The specified axis (``axis``) determines where the split occurs, and all other dimensions
        remain unchanged. See the example section below for the most common use cases.

        Parameters
        ----------
        x :
            The input array to be split, representing concatenated features, coefficients,
            or other data. The shape of ``x`` along the specified axis must match the total
            number of features generated by the basis, i.e., ``self.n_output_features``.

            **Examples:**
            - For a design matrix: ``(n_samples, total_n_features)``
            - For model coefficients: ``(total_n_features,)`` or ``(total_n_features, n_neurons)``.

        axis : int, optional
            The axis along which to split the features. Defaults to 1.
            Use ``axis=1`` for design matrices (features along columns) and ``axis=0`` for
            coefficient arrays (features along rows). All other dimensions are preserved.

        Raises
        ------
        ValueError
            If the shape of ``x`` along the specified axis does not match ``self.n_output_features``.

        Returns
        -------
        dict
            A dictionary where:

            - **Keys**: Labels of the additive basis components.
            - **Values**: Sub-arrays corresponding to each component. Each sub-array has the shape:

            .. math::
                (..., n_i1, n_i2,..,n_im, b_i, ...)

            - ``(n_samples, n_i1, n_i2,..,n_im)``: is the shape of the input to the i-th basis.
            - ``b_i``: The number of basis functions for the i-th basis component.

            These sub-arrays are reshaped along the specified axis, with all other dimensions
            remaining the same.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineConv
        >>> from nemos.glm import GLM
        >>> # Define an additive basis
        >>> basis = (
        ...     BSplineConv(n_basis_funcs=5, window_size=10, label="feature_1") +
        ...     BSplineConv(n_basis_funcs=6, window_size=10, label="feature_2")
        ... )
        >>> # Generate a sample input array and compute features
        >>> x1, x2 = np.random.randn(20), np.random.randn(20)
        >>> X = basis.compute_features(x1, x2)
        >>> # Split the feature matrix along axis 1
        >>> split_features = basis.split_by_feature(X, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        feature_1: shape (20, 5)
        feature_2: shape (20, 6)
        >>> # If one of the basis components accepts multiple inputs, the resulting dictionary will be nested:
        >>> multi_input_basis = BSplineConv(n_basis_funcs=6, window_size=10,
        ... label="multi_input")
        >>> X_multi = multi_input_basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = multi_input_basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        multi_input, shape (20, 2, 6)
        >>> # the method can be used to decompose the glm coefficients in the various features
        >>> counts = np.random.poisson(size=20)
        >>> model = GLM().fit(X, counts)
        >>> split_coef = basis.split_by_feature(model.coef_, axis=0)
        >>> for feature, coef in split_coef.items():
        ...     print(f"{feature}: shape {coef.shape}")
        feature_1: shape (5,)
        feature_2: shape (6,)

        """
        return super().split_by_feature(x, axis=axis)

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        The i-th axis of the grid will be sampled with ``n_samples[i]`` equi-spaced points.
        The method uses numpy.meshgrid with ``indexing="ij"``, returning matrix indexing
        instead of the default cartesian indexing, see Notes.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of points in the uniformly spaced grid. The length of
            n_samples must equal the number of combined bases.

        Returns
        -------
        *Xs :
            A tuple of arrays containing the meshgrid values, one element for each of the n dimension of the grid,
            where n equals to the number of inputs.
            The size of ``Xs[i]`` is ``(n_samples[0], ... , n_samples[n])``.
        Y :
            The basis function evaluated at the samples,
            shape ``(n_samples[0], ... , n_samples[n], number of basis)``.

        Raises
        ------
        ValueError
            If the time point number is inconsistent between inputs or if the number of inputs doesn't match what
            the Basis object requires.
        ValueError
            If one of the n_samples is <= 0.

        Notes
        -----
        Setting ``indexing = 'ij'`` returns a meshgrid with matrix indexing. In the N-D case with inputs of size
        :math:`M_1,...,M_N`, outputs are of shape :math:`(M_1, M_2, M_3, ....,M_N)`.
        This differs from the numpy.meshgrid default, which uses Cartesian indexing.
        For the same input, Cartesian indexing would return an output of shape :math:`(M_2, M_1, M_3, ....,M_N)`.

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define two basis objects and add them
        >>> basis_1 = nmo.basis.BSplineEval(10)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
        >>> additive_basis = basis_1 + basis_2

        >>> # evaluate on a grid of 10 x 10 equi-spaced samples
        >>> X, Y, Z = additive_basis.evaluate_on_grid(10, 10)

        """
        return super().evaluate_on_grid(*n_samples)

    def _get_feature_slicing(
        self,
        n_inputs: Optional[tuple] = None,
        start_slice: Optional[int] = None,
    ) -> Tuple[dict, int]:
        """
        Calculate and return the slicing for features based on the input structure.

        This method determines how to slice the features for different basis types.

        Parameters
        ----------
        n_inputs :
            The number of input basis for each component, by default it uses ``self._n_basis_input``.
        start_slice :
            The starting index for slicing, by default it starts from 0.

        Returns
        -------
        split_dict :
            Dictionary with keys as labels and values as slices representing
            the slicing for each additive component.
        start_slice :
            The updated starting index after slicing.

        See Also
        --------
        _get_default_slicing : Handles default slicing logic.
        _merge_slicing_dicts : Merges multiple slicing dictionaries, handling keys conflicts.
        """
        # Set default values for n_inputs and start_slice if not provided
        n_inputs = n_inputs or self._input_shape_product
        start_slice = start_slice or 0

        # If the instance is of AdditiveBasis type, handle slicing for the additive components

        split_dict, start_slice = self.basis1._get_feature_slicing(
            n_inputs[: len(self.basis1._input_shape_product)],
            start_slice,
        )
        sp2, start_slice = self.basis2._get_feature_slicing(
            n_inputs[len(self.basis1._input_shape_product) :],
            start_slice,
        )
        # label should always be unique, so update is safe
        split_dict.update(sp2)
        return split_dict, start_slice

    def __iter__(self):
        """Iterate over components."""
        for bas in self.basis1:
            yield bas
        for bas in self.basis2:
            yield bas

    def __len__(self):
        """Return the number of additive basis."""
        return len(self.basis1) + len(self.basis2)


class MultiplicativeBasis(CompositeBasisMixin, Basis):
    """
    Class representing the multiplication (external product) of two Basis objects.

    Parameters
    ----------
    basis1 :
        First basis object to multiply.
    basis2 :
        Second basis object to multiply.

    Examples
    --------
    >>> # Generate sample data
    >>> import numpy as np
    >>> import nemos as nmo
    >>> X = np.random.normal(size=(30, 3))

    >>> # define two basis and multiply
    >>> basis_1 = nmo.basis.BSplineEval(10)
    >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
    >>> multiplicative_basis = basis_1 * basis_2
    >>> multiplicative_basis
    '(BSplineEval * RaisedCosineLinearEval)': MultiplicativeBasis(
        ...
    )

    >>> # Can multiply or add another basis to the AdditiveBasis object
    >>> # This will cause the number of output features of the result basis to grow accordingly
    >>> basis_3 = nmo.basis.RaisedCosineLogEval(100)
    >>> multiplicative_basis_2 = multiplicative_basis * basis_3
    >>> multiplicative_basis_2
    '((BSplineEval * RaisedCosineLinearEval) * RaisedCosineLogEval)': MultiplicativeBasis(
        ...
    )
    """

    def __init__(
        self, basis1: BasisMixin, basis2: BasisMixin, label: Optional[str] = None
    ) -> None:
        if getattr(basis1, "is_complex", False) and getattr(
            basis2, "is_complex", False
        ):
            raise ValueError(
                "Invalid multiplication between two complex bases.\n"
                "Fourier basis are complex bases, and "
                "the multiplication of a real basis with a Fourier bases results in a complex basis as well. "
                "Multiplication between two complex bases is not allowed in NeMoS as it would treat "
                "real and imaginary columns alike."
            )
        input_shape1 = get_input_shape(basis1)
        input_shape2 = get_input_shape(basis2)
        # replace None with default
        input_shape1 = [() if i is None else i for i in input_shape1]
        input_shape2 = [() if i is None else i for i in input_shape2]
        have_unique_shapes, _, _ = _have_unique_shapes(input_shape1 + input_shape2)

        if not have_unique_shapes:
            basis1 = set_input_shape(deepcopy(basis1), None)
            basis2 = set_input_shape(deepcopy(basis2), None)
            warnings.warn(
                category=UserWarning,
                message="Multiple different input shapes detected. "
                "Resetting input shape to default (None).",
            )

        with self._set_shallow_copy(not have_unique_shapes):
            CompositeBasisMixin.__init__(self, basis1, basis2, label=label)
        Basis.__init__(self)

    def _generate_label(self) -> str:
        return "(" + self.basis1.label + " * " + self.basis2.label + ")"

    @property
    def n_basis_funcs(self):
        """Compute the n-basis function runtime.

        This plays well with cross-validation where the number of basis function of the
        underlying bases can be changed. It must be read-only since the number of basis
        is determined by the two basis elements and the type of composition.
        """
        return self.basis1.n_basis_funcs * self.basis2.n_basis_funcs

    @property
    def n_output_features(self):
        """Return the number of output features."""
        input_shape = self._input_shape_  # returns a list of length 1 or None
        if input_shape is None or input_shape[0] is None:
            return None
        n_basis1 = getattr(self.basis1, "n_basis_funcs")
        n_basis2 = getattr(self.basis2, "n_basis_funcs")
        return n_basis1 * n_basis2 * math.prod(input_shape[0])

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Evaluate the basis at the sample points.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs)

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(5) * nmo.basis.RaisedCosineLinearEval(6)
        >>> x, y = np.random.randn(2, 30)
        >>> X = mult_basis.evaluate(x, y)
        """
        # evaluate preserves the shape of the input arrays
        shape = xi[0].shape
        x1 = self.basis1.evaluate(*xi[: self.basis1._n_input_dimensionality])
        x2 = self.basis2.evaluate(*xi[self.basis1._n_input_dimensionality :])
        X = np.asarray(
            row_wise_kron(
                x1.reshape(-1, x1.shape[-1]),
                x2.reshape(-1, x2.shape[-1]),
                transpose=False,
            )
        )
        # run by the assumption (which is enforced) that the input shape is
        # shared in a multiplicative basis.
        X = X.reshape(*shape, -1)
        return X

    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Compute the features for the multiplied bases, and compute their outer product.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The  features, shape (n_samples, n_basis_funcs)

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(5) * nmo.basis.RaisedCosineLinearEval(6)
        >>> x, y = np.random.randn(2, 30)
        >>> X = mult_basis.compute_features(x, y)
        """
        kron = support_pynapple(conv_type="numpy")(row_wise_kron)
        comp_feature_1 = getattr(
            self.basis1, "_compute_features", self.basis1.compute_features
        )
        comp_feature_2 = getattr(
            self.basis2, "_compute_features", self.basis2.compute_features
        )
        x1 = comp_feature_1(*xi[: self.basis1._n_input_dimensionality])
        x2 = comp_feature_2(*xi[self.basis1._n_input_dimensionality :])
        # multiplicative basis inputs are of the same shape, checked and
        # set just before the call to this method
        n_samples = x1.shape[0]
        # flatten on the first axis, so that the rowise kron applies to
        # each sample and vectorized dimension in a pairwise way

        n_vec_inputs = self._input_shape_product[0]
        # note that x1/2 can have shape that doesn't divide self.basis1/2.n_basis_funcs
        # for example, OrthExponentialEval has an orthogonalization procedure that drops
        # redundant columns: i.e. the output may be less the nominal n-basis funcs,
        X = kron(
            x1.reshape((n_vec_inputs * n_samples, -1)),
            x2.reshape((n_vec_inputs * n_samples, -1)),
            transpose=False,
        )
        return X.reshape((n_samples, -1))

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        The i-th axis of the grid will be sampled with ``n_samples[i]`` equi-spaced points.
        The method uses numpy.meshgrid with ``indexing="ij"``, returning matrix indexing
        instead of the default cartesian indexing, see Notes.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of points in the uniformly spaced grid. The length of
            n_samples must equal the number of combined bases.

        Returns
        -------
        *Xs :
            A tuple of arrays containing the meshgrid values, one element for each of the n dimension of the grid,
            where n equals to the number of inputs.
            The size of ``Xs[i]`` is ``(n_samples[0], ... , n_samples[n])``.
        Y :
            The basis function evaluated at the samples,
            shape ``(n_samples[0], ... , n_samples[n], number of basis)``.

        Raises
        ------
        ValueError
            If the time point number is inconsistent between inputs or if the number of inputs doesn't match what
            the Basis object requires.
        ValueError
            If one of the n_samples is <= 0.

        Notes
        -----
        Setting ``indexing = 'ij'`` returns a meshgrid with matrix indexing. In the N-D case with inputs of size
        :math:`M_1,...,M_N`, outputs are of shape :math:`(M_1, M_2, M_3, ....,M_N)`.
        This differs from the numpy.meshgrid default, which uses Cartesian indexing.
        For the same input, Cartesian indexing would return an output of shape :math:`(M_2, M_1, M_3, ....,M_N)`.

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(4) * nmo.basis.RaisedCosineLinearEval(5)
        >>> X, Y, Z = mult_basis.evaluate_on_grid(10, 10)
        """
        return super().evaluate_on_grid(*n_samples)

    @add_docstring("compute_features", Basis)
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_mul = basis1 * basis2
        >>> X_multi = basis_mul.compute_features(np.random.randn(20, 2), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 60 = 5 * 2 * 6
        (20, 60)

        """
        return super().compute_features(*xi)

    @add_docstring("split_by_feature", BasisMixin)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_mul = basis1 * basis2
        >>> X_multi = basis_mul.compute_features(np.random.randn(20, 2), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 20 = 2 * 5 * 6
        (20, 60)

        >>> # The multiplicative basis is a single 2D component.
        >>> split_features = basis_mul.split_by_feature(X_multi, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        (one_input * two_inputs): shape (20, 2, 30)

        """
        # ruff: noqa: D205, D400
        return super().split_by_feature(x, axis=axis)

    def set_input_shape(self, *xi: int | tuple[int, ...] | NDArray) -> Basis:
        """
        Set the expected input shape for the basis object.

        This method sets the input shape for each component basis in the basis.
        One ``xi`` must be provided for each basis component, specified as an integer,
        a tuple of integers, or an array. The method calculates and stores the total number of output features
        based on the number of basis functions in each component and the provided input shapes.

        Parameters
        ----------
        *xi :
            The input shape specifications. For every k,``xi[k]`` can be:
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided, and it contains non-integer elements.
            If not enough inputs are provided.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.

        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define an additive basis
        >>> basis_1 = nmo.basis.BSplineEval(5)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(6)
        >>> basis_3 = nmo.basis.MSplineEval(7)
        >>> multiplicative_basis = basis_1 * basis_2 * basis_3

        Specify the input shape using all 3 allowed ways: integer, tuple, array
        >>> _ = multiplicative_basis.set_input_shape((4, 5), (4, 5), np.ones((10, 4, 5)))

        Expected output features are:
        (5 * 6 * 7 bases) * (20 inputs) = 4200
        >>> multiplicative_basis.n_output_features
        4200

        """
        # ruff: noqa: D400, D205
        super().set_input_shape(*xi, allow_inputs_of_different_shape=False)
        return self

    @property
    def _input_shape_(self):
        """Input shape list.

        Override default list by returning the shape of one of the inputs.
        Note that:
        1. The other inputs must have the same shape.
        2. This is used internally by `split_by_feature()` to know how many
           inputs are available.
        """
        return get_input_shape(self)[:1]
